# 786
# Aldy source: diplotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Tuple, Dict

import collections

from .common import allele_sort_key, td
from .gene import Gene, Mutation
from .coverage import Coverage
from .solutions import MinorSolution
from .version import __version__ as version


OUTPUT_COLS = [
    "Sample",
    "Gene",
    "SolutionID",
    "Major",
    "Minor",
    "Copy",
    "Allele",
    "Location",
    "Type",
    "Coverage",
    "Effect",
    "dbSNP",
    "Code",
    "Status",
]
"""list[str]: Output column descriptions"""


def write_decomposition(
    sample: str, gene: Gene, sol_id: int, minor: MinorSolution, f
) -> None:
    """
    Write an allelic decomposition to the file `f`.

    Args:
        sample (str): Sample name.
        gene (:obj:`aldy.gene.Gene`): Gene instance.
        sol_id (int): Solution ID (each solution must have a different ID).
        minor (:obj:`aldy.solutions.MinorSolution`):
            Minor star-allele solution to be written.
        f (file): Output file.
    """

    for copy, a in enumerate(minor.solution):
        mutations = set(gene.alleles[a.major].func_muts) | set(
            gene.alleles[a.major].minors[a.minor].neutral_muts
        )
        mutations |= set(a.added)
        mutations -= set(a.missing)
        items = []
        if len(mutations) > 0:
            for m in sorted(mutations):
                items.append(
                    [
                        sample,
                        gene.name,
                        sol_id,
                        minor.diplotype,
                        ";".join(ay.minor for ay in minor.solution),
                        copy,
                        a.minor,
                        m.pos,
                        m.op,
                        -1,
                        ["NEUTRAL", "DISRUPTING"][bool(m.is_functional)],
                        m.aux.get("dbsnp", ""),
                        m.aux.get("old", ""),
                        "",
                    ]
                )
        else:
            items.append(
                [
                    sample,
                    gene.name,
                    sol_id,
                    minor.diplotype,
                    ";".join(ay.minor for ay in minor.solution),
                    copy,
                    a.minor,
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                ]
            )
        for it in items:
            print("\t".join(str(i) for i in it), file=f)


def write_vcf(
    sample: str, gene: Gene, coverage: Coverage, minors: List[MinorSolution], f
):
    header = f"""
    ##fileformat=VCFv4.2
    ##source=aldy-v{version}
    ##INFO=<ID=ANN,Number=1,Type=String,Description="Location within {gene.name}">
    ##INFO=<ID=TYPE,Number=1,Type=String,Description="Mutation kind">
    ##INFO=<ID=TYPE,Number=1,Type=String,Description="Gene">
    ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
    ##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
    ##FORMAT=<ID=MA,Number=1,Type=String,Description="Major genotype star-allele calls">
    ##FORMAT=<ID=MI,Number=1,Type=String,Description="Minor genotype star-allele calls">
    """
    all_mutations: Dict[Mutation, List[dict]] = {
        m: [collections.defaultdict(int)] * len(minors)
        for minor in minors
        for a in minor.solution
        for m in set(gene.alleles[a.major].func_muts)
        | set(gene.alleles[a.major].minors[a.minor].neutral_muts)
        | set(a.added)
    }
    for mi, minor in enumerate(minors):
        for ai, a in enumerate(minor.solution):
            mutations = set(gene.alleles[a.major].func_muts) | set(
                gene.alleles[a.major].minors[a.minor].neutral_muts
            )
            mutations |= set(a.added)
            for m in mutations:
                all_mutations[m][mi][ai] = 1
    print(
        td(header).strip()
        + "\n"
        + "\t".join(
            ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT"]
            + [f"{sample}:{mi}:{m.diplotype}" for mi, m in enumerate(minors)]
        ),
        file=f,
    )
    for m in sorted(mutations):
        ref = gene.seq[m.pos - gene.region.start]
        if m.op[:3] == "SNP":
            alt = m.op[5]
        elif m.op[:3] == "INS":
            alt = ref + m.op[4:]
        else:
            ref = gene.seq[m.pos - 1 - gene.region.start]
            alt = ref + m.op[4:], ref

        info = [
            "ANN={}".format(m.aux.get("old", ".")),
            "TYPE={}".format(["NEUTRAL", "DISRUPTING"][bool(m.is_functional)]),
            "GENE=" + gene.name,
        ]
        data = []
        for mi, minor in enumerate(minors):
            nall = len(minor.solution)
            data.append(
                {
                    "GT": "|".join(str(all_mutations[m][mi][i]) for i in range(nall)),
                    "DP": str(coverage[m]),
                    "MA": ",".join(
                        f"*{minor.solution[i].major}"
                        if all_mutations[m][mi][i] > 0
                        else "-"
                        for i in range(nall)
                    ),
                    "MI": ",".join(
                        f"*{minor.solution[i].minor}"
                        if all_mutations[m][mi][i] > 0
                        else "-"
                        for i in range(nall)
                    ),
                }
            )
        pattern = (
            "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t"
            + "{format}\t{data}"
        )
        print(
            pattern.format(
                chrom=gene.region.chr,
                pos=m.pos + 1,
                id=m.aux.get("dbsnp", "."),
                ref=ref,
                alt=alt,
                qual=0,
                filter="PASS",
                info=";".join(info),
                format=":".join(data[0].keys()),
                data="\t".join(":".join(d.values()) for d in data),
            ),
            file=f,
        )


def estimate_diplotype(gene: Gene, solution: MinorSolution) -> str:
    """
    Calculate the diplotype assignment for a minor solution.
    Assigns a ``diplotype`` attribute of the :obj:`aldy.solutions.MinorSolution`.
    Relies on the diplotype assignment heuristics to assign correct diplotypes.
    This heuristics has no biological validity whatsoever--- it is purely used
    to pretty-print the final solutions.

    Returns:
        str: Diplotype assignment.
    """

    del_allele = gene.deletion_allele()

    # solution is the array of (major, minor) tuples
    majors = [
        str(allele_sort_key(a.major)[0])
        + ("-like" if sum(1 for m in a.added if m.is_functional) > 0 else "")
        for a in solution.solution
    ]
    diplotype: Tuple[List[str], List[str]] = ([], [])

    if len(majors) == 1 and del_allele:
        majors.append(del_allele)

    major_dict = collections.Counter(majors)
    dc = 0

    # Handle tandems (heuristic that groups common tandems together,
    #                 e.g. 1, 2, 13 -> 1+13/2 if [1,13] is a common tandem)
    for ta, tb in gene.common_tandems:
        while major_dict[ta] > 0 and major_dict[tb] > 0:
            diplotype[dc % 2].extend([ta, tb])
            dc += 1
            major_dict[ta] -= 1
            major_dict[tb] -= 1

    # Handle duplicates (heuristics that groups duplicate alleles together,
    #                    e.g. 1, 1, 2 -> 1+1/2)
    # First check should we split them (e.g. 1, 1, 1, 1 -> 1+1/1+1)?
    el = list(major_dict.elements())
    if len(major_dict) == 1 and len(el) % 2 == 0:
        p, a = len(el) // 2, el[0]
        diplotype = (p * [a], p * [a])
        major_dict = collections.Counter()
    for allele, count in major_dict.items():
        if count > 1:
            if len(diplotype[dc % 2]) > len(diplotype[(dc + 1) % 2]):
                dc += 1
            diplotype[dc % 2].extend(count * [str(allele)])
            major_dict[allele] -= count
            dc += 1

    # Handle the rest
    for allele, count in major_dict.items():
        if count > 0:
            if len(diplotype[dc % 2]) > len(diplotype[(dc + 1) % 2]):
                dc += 1
            assert count == 1
            diplotype[dc % 2].append(str(allele))
            dc += 1

    # Each diplotype should have at least one item
    # e.g. 1, 1 -> becomes 1+1/_ due to duplicate heuristic -> fixed to 1/1
    if len(diplotype[1]) == 0:
        diplotype = (diplotype[0][:-1], [diplotype[0][-1]])

    # Make sure that the elements are sorted and that the tandems are grouped together
    result = sorted(diplotype)
    for i in range(2):
        nd: List[tuple] = []
        for ta, tb in gene.common_tandems:
            if ta in result[i] and tb in result[i]:
                nd.append((ta, tb))
                result[i].remove(ta)
                result[i].remove(tb)
        nd += [(x,) for x in result[i]]
        result[i] = [f for e in sorted(nd) for f in e]

    res = "/".join("+".join("*{}".format(y) for y in x) for x in result)
    solution.diplotype = res
    return res



# 786
# Aldy source: test_diplotype_real.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest  # noqa

from .test_diplotype_synthetic import assert_diplotype


def test_basic(real_gene):
    assert_diplotype(real_gene, "*1/*1", ["1", "1.a"])


def test_tandem(real_gene):
    assert_diplotype(real_gene, "*2+*2/*71", ["2", "2.a", "71"])
    assert_diplotype(real_gene, "*4+*4/*41", ["4.b", "41", "4.b"])
    assert_diplotype(real_gene, "*4/*4+*4", ["4.b", "4", "4.b"])
    assert_diplotype(real_gene, "*3/*68+*4", ["3", "4", "68"])


def test_36(real_gene):
    assert_diplotype(real_gene, "*36+*10/*36+*41", ["10", "36", "41", "36"])
    assert_diplotype(real_gene, "*1+*36/*36+*10", ["1", "10", "36", "36"])
    assert_diplotype(real_gene, "*10/*36+*10", ["10", "36", "10"])
    assert_diplotype(real_gene, "*36+*10/*36+*10", ["10", "36", "36", "10"])
    assert_diplotype(real_gene, "*10+*10/*36+*10", ["10", "36", "10", "10"])


def test_fusion(real_gene):
    assert_diplotype(real_gene, "*2+*2/*68+*4", ["2", "4", "68", "2"])
    assert_diplotype(real_gene, "*1/*79+*2", ["1", "2", "79/2"])
    assert_diplotype(real_gene, "*2/*78+*2", ["2", "78/2", "2"])
    assert_diplotype(real_gene, "*1/*78+*2", ["2", "1", "78/2"])



# 786
# Aldy source: test_minor_synthetic.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest  # noqa
import collections

from aldy.solutions import CNSolution, SolvedAllele, MajorSolution
from aldy.coverage import Coverage
from aldy.gene import Mutation
from aldy.major import NOVEL_MUTATION_PENAL
from aldy.minor import estimate_minor, ADD_PENALTY_FACTOR, MISS_PENALTY_FACTOR
from aldy.common import SOLUTION_PRECISION, sorted_tuple


def assert_minor(gene, solver, data, shallow=False):
    cn_sol = CNSolution(0, list(collections.Counter(data["cn"]).elements()), gene)

    cov = collections.defaultdict(dict)
    for (pos, op), c in data["data"].items():
        cov[pos][op] = c
    cov = Coverage(cov, 0.5, {})

    major_solved = {
        SolvedAllele(maj, None, tuple(), tuple())
        if not isinstance(maj, tuple)
        else SolvedAllele(
            maj[0], None, tuple(Mutation(mp, mo, True) for mp, mo in maj[1:]), tuple()
        ): cnt
        for maj, cnt in data["major"].items()
    }
    major = MajorSolution(0, major_solved, cn_sol)
    sols = estimate_minor(gene, cov, [major], solver)

    if "score" in data:
        for s in sols:
            assert abs(data["score"] - s.score) < SOLUTION_PRECISION, "Score"

    if not shallow:
        sols_expected = [
            sorted_tuple(
                (i[0], sorted_tuple(i[1]), sorted_tuple(i[2])) for i in data["sol"]
            )
        ]
        sols_parsed = [
            sorted_tuple(
                (
                    i.minor,
                    sorted_tuple((m.pos, m.op) for m in i.missing),
                    sorted_tuple((m.pos, m.op) for m in i.added),
                )
                for i in s.solution
            )
            for s in sols
        ]
        assert sorted(sols_expected) == sorted(sols_parsed)
    else:
        # As assignments can vary between multiple optimal solutions,
        # just test minor allele assignments
        eall, emiss, enew = set(), set(), set()
        for i in data["sol"]:
            eall.add(i[0])
            emiss |= set(i[1])
            enew |= set(i[2])
        assert len(sols) == 1, "Single minor solution"
        pall, pmiss, pnew = set(), set(), set()
        for i in sols[0].solution:
            pall.add(i.minor)
            pmiss |= set((m.pos, m.op) for m in i.missing)
            pnew |= set((m.pos, m.op) for m in i.added)

        assert eall == pall, "Alleles"
        assert emiss == pmiss, "Missing mutations"
        assert enew == pnew, "Novel mutations"
    return sols[0].score if len(sols) > 0 else 0


def test_basic(toy_gene, solver):
    # Test two copies of *1
    assert_minor(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {(115, "_"): 20},
            "major": {"1": 2},
            "sol": [("1", [], []), ("1", [], [])],
            "score": 0,
        },
    )


def test_minor(toy_gene, solver):
    assert_minor(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {(115, "_"): 9, (115, "SNP.TA"): 11},
            "major": {"1": 2},
            "sol": [("1", [], []), ("1B", [], [])],
            "score": 0.2,
        },
    )


def test_miss(toy_gene, solver):
    assert_minor(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {
                (115, "_"): 10,
                (115, "SNP.TA"): 10,
                (148, "_"): 20,
                (151, "_"): 10,
                (151, "SNP.CT"): 10,
            },
            "major": {"1": 1, "3": 1},
            "sol": [("1B", [], []), ("3", [(148, "INS.A")], [])],
            "score": MISS_PENALTY_FACTOR,
        },
    )


def test_add(toy_gene, solver):
    assert_minor(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {
                (115, "_"): 0,
                (115, "SNP.TA"): 20,
                (148, "_"): 20,
                (148, "INS.A"): 10,
                (151, "_"): 10,
                (151, "SNP.CT"): 10,
            },
            "major": {"1": 1, "3": 1},
            "sol": [("1B", [], []), ("3", [], [(115, "SNP.TA")])],
            "score": ADD_PENALTY_FACTOR,
        },
    )


def test_major_novel(toy_gene, solver):
    assert_minor(
        toy_gene,
        solver,
        {
            "cn": {"1": 2, "6": 1},
            "data": {
                (111, "_"): 10,
                (111, "DEL.AC"): 10,
                (115, "_"): 10,
                (115, "SNP.TA"): 10,
                (148, "_"): 20,
                (151, "_"): 10,
                (151, "SNP.CT"): 10,
            },
            "major": {("1", (111, "DEL.AC")): 1, "3": 1, "6": 1},
            "sol": [
                ("1B", [], [(111, "DEL.AC")]),
                ("3", [(148, "INS.A")], []),
                ("6DEL", [], []),
            ],
            "score": MISS_PENALTY_FACTOR + NOVEL_MUTATION_PENAL,
        },
    )



# 786
# Aldy source: solutions.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, List

import collections

from .common import allele_sort_key
from .gene import Gene, GeneRegion


class CNSolution(
    collections.namedtuple("CNSolution", ["score", "solution", "gene", "region_cn"])
):
    """
    Valid copy-number configuration assignment.
    Immutable.

    Attributes:
        score (float):
            ILP model objective score (0 for user-provided solutions).
        solution (dict[str, int]):
            Copy-number configurations mapped to the corresponding copy number
            (e.g. ``{1: 2}`` means that there are two copies of \*1 configuration).
        region_cn (dict[:obj:`aldy.common.GeneRegion`, int]):
            Gene region copy numbers inferred by this solution.
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.

    Notes:
        Has custom printer (``__str__``).
    """

    def __new__(self, score: float, solution: List[str], gene: Gene):
        vec: Dict[int, Dict[GeneRegion, float]] = collections.defaultdict(
            lambda: collections.defaultdict(int)
        )
        for conf in solution:
            for g in gene.cn_configs[conf].cn:
                for r in gene.cn_configs[conf].cn[g]:
                    vec[g][r] += gene.cn_configs[conf].cn[g][r]
        return super(CNSolution, self).__new__(  # type:ignore
            self,
            score,
            collections.Counter(solution),
            gene,
            {a: dict(b) for a, b in vec.items()},
        )

    def position_cn(self, pos: int) -> float:
        """
        Returns:
            float: Copy number at the locus ``pos``.
        """
        try:
            g, region = self.gene.region_at(pos)
            return self.region_cn[g][region]
        except KeyError:
            return 0

    def _solution_nice(self):
        return ",".join(
            f"{v}x*{k}"
            for k, v in sorted(
                self.solution.items(), key=lambda x: allele_sort_key(x[0])
            )
        )

    def __str__(self):
        regions = sorted(set(r for g in self.region_cn for r in self.region_cn[g]))
        return "CNSol[{:.2f}; sol=({}); cn={}]".format(
            self.score,
            self._solution_nice(),
            "|".join(
                "".join(
                    "{:.0f}".format(self.region_cn[g][r])
                    if r in self.region_cn[g]
                    else "_"
                    for r in regions
                )
                for g in sorted(self.region_cn)
            ),
        )


class SolvedAllele(
    collections.namedtuple("SolvedAllele", ["major", "minor", "added", "missing"])
):
    """
    Solved star-allele assignment.
    Immutable.

    Attributes:
        major (str):
            Major star-allele identifier.
        minor (str, optional):
            Minor star-allele identifier. Can be None.
        added (tuple[:obj:`aldy.gene.Mutation`]):
            Mutations that are added to the star-allele
            (i.e. these mutations are not present in the allele database definition).
        missing (tuple[:obj:`aldy.gene.Mutation`]):
            Mutations that are omitted from the star-allele
            (i.e. these mutations are present in the allele database definition
            but not here).

    Notes:
        Has custom printer (``__str__``).
    """

    def major_repr(self):
        """
        Pretty-formats major star-allele name.
        """
        return "*{}{}".format(
            self.major,
            "".join(
                " +" + str(m) for m in sorted(m for m in self.added if m.is_functional)
            ),
        )

    def __str__(self):
        """
        Pretty-formats minor star-allele name.
        """
        return "*{}{}{}".format(
            self.minor if self.minor else self.major,
            "".join(
                " +" + str(m)
                for m in sorted(
                    self.added, key=lambda m: (-m.is_functional, m.pos, m.op)
                )
            ),
            "".join(" -" + str(m) for m in sorted(self.missing)),
        )


class MajorSolution(
    collections.namedtuple("MajorSolution", ["score", "solution", "cn_solution"])
):
    """
    Valid major star-allele assignment.
    Immutable.

    Attributes:
        score (float):
            ILP model objective score.
        solution (dict[:obj:`SolvedAllele`, int]):
            Major star-alleles and the corresponding copy numbers
            (e.g. ``{1: 2}`` means that we have two copies of \*1).
        cn_solution (:obj:`aldy.solutions.CNSolution`):
            Copy-number solution that was used to assign major star-alleles.

    Notes:
        Has custom printer (``__str__``).
    """

    def _solution_nice(self):
        return ", ".join(
            f"{v}x{s}"
            for s, v in sorted(
                self.solution.items(), key=lambda x: allele_sort_key(x[0].major)
            )
        )

    def __str__(self):
        return (
            f"MajorSol[{self.score:.2f}; "
            + f"sol=({self._solution_nice()}); "
            + f"cn={self.cn_solution}"
        )


class MinorSolution(
    collections.namedtuple("MinorSolution", ["score", "solution", "major_solution"])
):
    """
    Valid minor star-allele assignment.
    Immutable.

    Attributes:
        score (float):
            ILP model objective score.
        solution (list[:obj:`SolvedAllele`]):
            List of solved minor star-alleles.
            Modifications to the minor alleles are represented in
            :obj:`SolvedAllele` format.
        major_solution (:obj:`aldy.solutions.MajorSolution`):
            Major star-allele solution used for calculating the
            minor star-allele assignment.
        diplotype (str):
            Diplotype assignment (e.g. ``*1/*2``).

    Notes:
        Has custom printer (``__str__``).
    """

    diplotype = ""

    def _solution_nice(self):
        return ", ".join(
            str(s)
            for s in sorted(self.solution, key=lambda x: allele_sort_key(x.minor))
        )

    def __str__(self):
        return (
            f"MinorSol[{self.score:.2f}; "
            + f"sol=({self._solution_nice()}); "
            + f"major={self.major_solution}"
        )



# 786
# Aldy source: sam.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Tuple, Dict, List, Optional, Any

import pysam
import os
import re
import gzip
import struct
import collections

from .common import log, GRange, AldyException, script_path
from .coverage import Coverage
from .gene import Gene, Mutation


CIGAR2CODE: Dict[int, int] = {ord(y): x for x, y in enumerate("MIDNSHP=XB")}
"""str: CIGAR characters for an ASCII ordinal of CIGAR character"""

CIGAR_REGEX = re.compile(r"(\d+)([MIDNSHP=XB])")
"""regex: Regex that matches valid CIGAR strings"""


DEFAULT_CN_NEUTRAL_REGION = GRange("22", 42547463, 42548249)
"""obj:`aldy.common.GRange` Default copy-number neutral region
   (exon 4-6 of the CYP2D8 gene)"""


class Sample:
    """
    Interface for reading SAM/BAM/CRAM files that parses
    and stores read alignments.

    Attributes:
        coverage (:obj:`aldy.coverage.Coverage`):
            The coverage data for the sample.
            Consult the :obj:`aldy.coverage.Coverage`.
        reads (dict[str, tuple[:obj:`pysam.AlignedSegment`,
               :obj:`pysam.AlignedSegment`]]):
            Dictionary that keeps the paired-end reads together in a tuple.
            Key is the read name. Used only if ``phase`` is set.

    Static methods:
        ``load_sam_profile`` (documentation below).
    """

    def __init__(
        self,
        sam_path: str,
        gene: Gene,
        threshold: float,
        profile: str,
        phase: bool = False,
        reference: Optional[str] = None,
        cn_region: Optional[GRange] = DEFAULT_CN_NEUTRAL_REGION,
        debug: Optional[str] = None,
    ) -> None:
        """
        Initialize a :obj:`Sample` object.

        Args:
            sam_path (str):
                Path to a SAM/BAM/CRAM file.
            gene (:obj:`aldy.gene.Gene`):
                Gene instance.
            threshold (float):
                Threshold for filtering out low quality mutations. Ranges from 0 to 1.
                Check :obj:`aldy.coverage.Coverage` for more information.
            profile (str, optional):
                Profile specification (e.g. 'prgnseq-v1') or a profile SAM/BAM file.
            phase (bool):
                Construct basic rudimentary phasing of the reads to aid the genotyping.
                DEPRECATED: currently does nothing.
                Default is ``False``.
            reference (str, optional):
                Reference genome for reading CRAM files.
                Default is None.
            cn_region (:obj:`aldy.common.GRange`, optional):
                Copy-number neutral region to be used for coverage rescaling.
                If None, profile loading and coverage rescaling will be skipped
                (and Aldy will require a ``--cn`` parameter to be user-provided).
                Default is ``DEFAULT_CN_NEUTRAL_REGION``.
            debug (str, optional):
                If set, create a "`debug`.dump" file for debug purposes.
                Default is None.

        Raises:
            :obj:`aldy.common.AldyException` if the average coverage of the
            copy-number neutral region is too low (less than 2).
        """

        self.load_aligned(sam_path, gene, threshold, phase, reference, cn_region, debug)
        if cn_region:
            self.detect_cn(gene, profile, cn_region)
        log.debug("SAM: avg_coverage = {}", self.coverage.average_coverage())
        if cn_region and self.coverage.diploid_avg_coverage() < 2:
            raise AldyException(
                "The average coverage of the sample is too low ({:1f}).".format(
                    self.coverage.diploid_avg_coverage()
                )
            )

    def load_aligned(
        self,
        sam_path: str,
        gene: Gene,
        threshold: float,
        phase: bool = False,
        reference: Optional[str] = None,
        cn_region: Optional[GRange] = None,
        debug: Optional[str] = None,
    ) -> None:
        """
        Load the read, mutation and coverage data from a SAM/BAM/CRAM file.

        Args:
            sam_path (str):
                Path to a SAM/BAM/CRAM file.
            gene (:obj:`aldy.gene.Gene`):
                Gene instance.
            threshold (float):
                Threshold for filtering out low quality mutations. Ranges from 0 to 1.
                Check :obj:`aldy.coverage.Coverage` for more information.
            phase (bool):
                Construct basic rudimentary phasing of the reads to aid the genotyping.
                DEPRECATED: currently does nothing.
                Default is False.
            reference (str, optional):
                Reference genome for reading CRAM files.
                Default is None.
            cn_region (:obj:`aldy.common.GRange`, optional):
                Copy-number neutral region to be used for coverage rescaling.
                If None, profile loading and coverage rescaling will be skipped
                (and Aldy will require a ``--cn`` parameter to be user-provided).
                Default is ``DEFAULT_CN_NEUTRAL_REGION`` (CYP2D8).
            debug (str, optional):
                If set, create a "`debug`.dump" file for debug purposes.
                Default is None.

        Raises:
            :obj:`aldy.common.AldyException` if a BAM/CRAM file lacks an index.
        """

        # store paired-end reads (only if phasing is on)
        self.reads: Dict[str, Tuple[pysam.AlignedSegment, pysam.AlignedSegment]] = {}

        log.debug("SAM: path = {}", sam_path)

        # dict of int: int: the coverage of the CN-neutral region
        cnv_coverage: Dict[int, int] = collections.defaultdict(int)
        # Get the list of indel sites that should be corrected
        # TODO: currently uses only functional indels; other indels should be
        # corrected as well
        _indel_sites: Dict[Mutation, tuple] = {
            m: (collections.defaultdict(int), 0)
            for an, a in gene.alleles.items()
            for m in a.func_muts
            if m.op[:3] == "INS"
        }
        muts: dict = collections.defaultdict(int)
        norm: dict = collections.defaultdict(int)

        if sam_path[-5:] == ".dump":
            if phase:
                raise AldyException("Debug dumps do not work with --phase parameter")
            with gzip.open(sam_path, "rb") as fd:
                log.warn("Loading debug dump from {}", sam_path)
                l, h, i = (
                    struct.calcsize("<l"),
                    struct.calcsize("<h"),
                    struct.calcsize("<i"),
                )
                m, M = struct.unpack("<ll", fd.read(l + l))
                for j in range(m, M + 1):
                    cnv_coverage[j], = struct.unpack("<i", fd.read(i))
                ld, = struct.unpack("<l", fd.read(l))
                for _ in range(ld):
                    ref_start, ref_end, read_len, num_mutations = struct.unpack(
                        "<llhh", fd.read(l + l + h + h)
                    )
                    ref_end += ref_start
                    for j in range(ref_start, ref_end):
                        norm[j] += 1
                    insertions = set()
                    for __ in range(num_mutations):
                        mut_start, op_len = struct.unpack("<hh", fd.read(h + h))
                        mut_start += ref_start
                        op = fd.read(op_len).decode("ascii")
                        muts[mut_start, op] += 1
                        if op[:3] == "DEL":
                            for j in range(0, len(op) - 4):
                                norm[mut_start + j] -= 1
                        elif op[:3] == "INS":
                            insertions.add((mut_start, len(op) - 4))
                        else:
                            norm[mut_start] -= 1
                    for ins in _indel_sites:
                        ins_len = len(ins.op) - 4
                        if (ins.pos, ins_len) in insertions:
                            _indel_sites[ins] = (
                                _indel_sites[ins][0],
                                _indel_sites[ins][1] + 1,
                            )
                        else:
                            _indel_sites[ins][0][(ref_start, ref_end, read_len)] += 1
        else:
            with pysam.AlignmentFile(sam_path, reference_filename=reference) as sam:
                # Check do we have proper index to speed up the queries
                try:
                    sam.check_index()
                except AttributeError:
                    pass  # SAM files do not have an index. BAMs might also lack it
                except ValueError:
                    raise AldyException(
                        f"File {sam_path} has no index (it must be indexed)"
                    )

                #: str: Check if we need to append 'chr' or not
                self._prefix = _chr_prefix(
                    gene.region.chr, [x["SN"] for x in sam.header["SQ"]]
                )

                # Try to read CN-neutral region if there is an index (fast)
                # If not, the main loop will catch it later on
                def read_cn_read(read):
                    """
                    Check is pysam.AlignmentSegment valid CN-neutral read,
                    and if so, parse it
                    """
                    start = read.reference_start
                    if read.cigartuples is None or (read.flag & 0x40) == 0:
                        return
                    for op, size in read.cigartuples:
                        if op in [0, 7, 8, 2]:
                            for i in range(size):
                                cnv_coverage[start + i] += 1
                            start += size

                # Set it to _fetched_ if a CN-neutral region is user-provided.
                # Then read the CN-neutral region.
                is_cn_region_fetched = cn_region is None
                if cn_region and sam.has_index():
                    log.debug("File {} has index", sam_path)
                    for read in sam.fetch(
                        region=cn_region.samtools(prefix=self._prefix)
                    ):
                        read_cn_read(read)
                    is_cn_region_fetched = True

                # Fetch the reads
                total = 0
                dump_data = []
                for read in sam.fetch(region=gene.region.samtools(prefix=self._prefix)):
                    # If we haven't obtained CN-neutral region so far, do it now
                    if (
                        not is_cn_region_fetched
                        and cn_region
                        and _in_region(cn_region, read, self._prefix)
                    ):  # type: ignore
                        read_cn_read(read)

                    r = self._parse_read(
                        read, gene, norm, muts, _indel_sites, debug is not None
                    )
                    if r:
                        total += 1
                        if debug:
                            dump_data.append(r)
                        if phase:
                            if read.query_name in self.reads:
                                assert self.reads[read.query_name][1] is None
                                read1 = self.reads[read.query_name][0]
                                if read.reference_start < read1.reference_start:
                                    read, read1 = read1, read
                                self.reads[read.query_name] = (read1, read)
                            else:
                                self.reads[read.query_name] = (read, None)

                if debug:
                    if phase:
                        raise AldyException(
                            "Debug dumps do not work with --phase parameter"
                        )
                    with gzip.open(f"{debug}.dump", "wb") as fd:
                        m, M = min(cnv_coverage.keys()), max(cnv_coverage.keys())
                        fd.write(struct.pack("<ll", m, M))
                        for i in range(m, M + 1):
                            fd.write(struct.pack("<i", cnv_coverage[i]))
                        fd.write(struct.pack("<l", len(dump_data)))
                        for (s, e, l), m in dump_data:
                            fd.write(struct.pack("<llhh", s, e - s, l, len(m)))
                            for p, md in m:
                                fd.write(struct.pack("<hh", p - s, len(md)))
                                fd.write(md.encode("ascii"))

        # Establish the coverage dictionary
        coverage: Dict[int, Dict[str, int]] = dict()
        for pos, cov in norm.items():
            if cov == 0:
                continue
            if pos not in coverage:
                coverage[pos] = {}
            coverage[pos]["_"] = cov
        for m, cov in muts.items():
            pos, mut = m
            if pos not in coverage:
                coverage[pos] = {}
            coverage[pos][mut] = cov

        #: dict of int: (dict of str: int): coverage dictionary
        #: keys are loci in the reference genome, while values are dictionaries
        #: that describe the coverage of each mutation
        # (_ stands for non-mutated nucleotide)
        self.coverage = Coverage(coverage, threshold, cnv_coverage)
        for mut, ins_cov in _indel_sites.items():
            if mut.pos in coverage and mut.op in coverage[mut.pos]:
                self._correct_ins_coverage(mut, ins_cov)

    def _parse_read(
        self,
        read: pysam.AlignedSegment,
        gene: Gene,
        norm: dict,
        muts: dict,
        indel_sites=None,
        dump=False,
    ) -> Optional[Tuple[Tuple[int, int, int], Any]]:
        """
        Parse a :obj:`pysam.AlignedSegment` read.

        Params:
            read (:obj:`pysam.AlignedSegment`)
            gene (:obj:`aldy.gene.Gene`)

        Returns:
            optional: None if parsing was not successful.
            Otherwise, returns the tuple consisting of:

                - start and end positions of the read, and
                - the list of mutations found in the read.

        Params that are mutated:
            norm (:obj:`collections.defaultdict(int)`):
                dict of positions within the read that have not been mutated.
            muts (:obj:`collections.defaultdict(int)`):
                dict of positions within the read that have been mutated.
        """

        if not _in_region(
            gene.region, read, self._prefix
        ):  # ensure that it is a proper gene read
            return None
        if not read.cigartuples:  # only valid alignments
            return None
        if read.flag & 0x800:  # avoid supplementary alignments
            return None
        if "H" in read.cigarstring:  # avoid hard-clipped reads
            return None

        insertions = set()
        dump_arr = []
        start, s_start = read.reference_start, 0
        for op, size in read.cigartuples:
            if op == 2:  # Deletion
                mut = (
                    start,
                    "DEL.{}".format(
                        gene.seq[
                            start - gene.region.start : start - gene.region.start + size
                        ]
                    ),
                )
                muts[mut] += 1
                if dump:
                    dump_arr.append(mut)
                start += size
            elif op == 1:  # Insertion
                mut = (
                    start,
                    "INS.{}".format(
                        read.query_sequence[s_start : s_start + size].lower()
                    ),
                )
                muts[mut] += 1
                if dump:
                    dump_arr.append(mut)
                insertions.add((start, size))
                s_start += size
            elif op == 4:  # Soft-clip
                s_start += size
            elif op in [0, 7, 8]:  # M, X and =
                for i in range(size):
                    if not 0 <= start + i - gene.region.start < len(gene.seq):
                        continue
                    if (
                        gene.seq[start + i - gene.region.start]
                        != read.query_sequence[s_start + i]
                    ):
                        mut = (
                            start + i,
                            "SNP.{}{}".format(
                                gene.seq[start + i - gene.region.start],
                                read.query_sequence[s_start + i],
                            ),
                        )
                        if dump:
                            dump_arr.append(mut)
                        muts[mut] += 1
                    else:
                        norm[start + i] += 1
                start += size
                s_start += size

        if indel_sites is not None:
            for ins in indel_sites:
                ins_len = len(ins.op) - 4
                if (ins.pos, ins_len) in insertions:
                    indel_sites[ins] = (indel_sites[ins][0], indel_sites[ins][1] + 1)
                else:
                    indel_sites[ins][0][
                        (read.reference_start, start, len(read.query_sequence))
                    ] += 1
        return (read.reference_start, start, len(read.query_sequence)), dump_arr

    def _correct_ins_coverage(self, mut: Mutation, j) -> None:
        """
        Fix low coverage of large tandem insertions.

        Targets the cases where reference looks like
            ...X...
        and the donor genome looks like
            ...XX... (i.e. X is a tandem insertion).
        Any read that covers only one tandem (e.g. read X...)
        will get perfectly aligned (without trigerring an insertion tag)
        to the tail of the tandem insertion.
        This results in under-estimation of the insertion abundance will
        (as aligner could not assign `I` CIGAR to the correct insertion).
        This function attempts to correct this bias.

        Notes:
            This function modifies ``self.coverage``.
        """

        MARGIN_RATIO = 0.20
        INSERT_LEN_RESCALE = 10.0
        INSERT_LEN_AMPLIFY = 3

        total = self.coverage.total(mut.pos)
        current_ratio = float(self.coverage[mut]) / total

        j0 = 0
        for orig_start, start, read_len in j[0]:
            ins_len = len(mut.op) - 4
            # HACK: This is rather hackish way of detecting *40
            # min 20% on the sides, max ... how much?
            min_margin = MARGIN_RATIO * max(1, ins_len / INSERT_LEN_RESCALE) * read_len
            if start < mut.pos + max(
                INSERT_LEN_AMPLIFY * ins_len, min_margin
            ) or orig_start > mut.pos - max(
                (INSERT_LEN_AMPLIFY - 1) * ins_len, min_margin
            ):
                continue
            j0 += j[0][(orig_start, start, read_len)]

        new_ratio = j[1] / float(j0 + j[1])
        log.debug(
            "Rescaling indel: {}:{} from {}/{} to {} (indel:total = {}:{})",
            mut.pos,
            mut.op,
            self.coverage[mut],
            total,
            int(self.coverage[mut] * (new_ratio / current_ratio)),
            j[1],
            j0 + j[1],
        )
        self.coverage._coverage[mut.pos][mut.op] = int(
            self.coverage[mut] * (new_ratio / current_ratio)
        )

    # ----------------------------------------------------------------------------------
    # Coverage-rescaling functions
    # ----------------------------------------------------------------------------------

    def detect_cn(self, gene: Gene, profile: str, cn_region: GRange) -> None:
        """
        Rescale the ``self.coverage`` to fit the sequencing profile.

        Params:
            gene (:obj:`aldy.gene.Gene`):
                Gene instance.
            profile (str):
                Profile identifier (e.g. 'pgrnseq-v1'). Can be a SAM/BAM file as well.
            cn_region (:obj:`aldy.common.GRange`):
                Coordinates of the copy-number neutral region.

        Notes:
            This function assumes that ``self.coverage`` is set.
            It modifies ``self.coverage``.
        """

        if os.path.exists(profile):
            ext = os.path.splitext(profile)
            if ext in ["bam", "sam"]:
                prof = self._load_profile(
                    profile, is_bam=True, gene_region=gene.region, cn_region=cn_region
                )
            else:
                prof = self._load_profile(profile)
        else:
            profile_path = script_path(
                "aldy.resources.profiles/{}.profile".format(profile.lower())
            )
            prof = self._load_profile(profile_path)
        self.coverage._normalize_coverage(prof, gene.regions, cn_region)

    def _load_profile(
        self,
        profile_path: str,
        is_bam: bool = False,
        gene_region: Optional[GRange] = None,
        cn_region: Optional[GRange] = None,
    ) -> Dict[str, Dict[int, float]]:
        """
        Load a coverage profile.

        Returns:
            defaultdict[str, dict[int, float]]: A profile dictionary where
            keys are chromosome IDs (e.g. '7' for chr7)
            and values are dictionaries that map the genomic locus to the
            corresponding profile coverage.
            This is a :obj:`collections.defaultdict` that uses 0 as a placeholder
            for the missing loci.

        Args:
            profile_path:
                Path to a profile file.
                SAM/BAM files are also accepted if `is_bam` is set
                (profile will be dynamically calculated in that case).
            is_bam (bool):
                A flag indicating if the `profile_path` is SAM/BAM or not.
            gene_region (:obj:`aldy.common.GRange`, optional):
                Region to be extracted from the profile.
                Default is None.
            cn_region (:obj:`aldy.common.GRange`, optional):
                Copy-number neutral region to be extracted from the profile.
                Default is None.

        Raises:
            :obj:`aldy.common.AldyException` if ``is_bam`` is set but ``gene_region``
            and ``cn_region`` are not.
        """

        profile: Dict[str, Dict[int, float]] = collections.defaultdict(
            lambda: collections.defaultdict(int)
        )
        if is_bam:
            if gene_region and cn_region:
                ptr = Sample.load_sam_profile(
                    profile_path, regions=[gene_region, cn_region]
                )
                for _, c, p, v in ptr:
                    profile[c][p] = v
            else:
                raise AldyException("Region parameters must be provided")
        else:
            with open(profile_path) as f:
                for line in f:
                    if line[0] == "#":
                        continue  # skip comments
                    ch, pos, val = line.strip().split()[1:]
                    profile[ch][int(pos)] = float(val)
        return profile

    # ----------------------------------------------------------------------------------

    @staticmethod
    def load_sam_profile(
        sam_path: str, factor: float = 2.0, regions: Optional[List[GRange]] = None
    ) -> List[Tuple[str, str, int, float]]:
        """
        Load the profile information from a SAM/BAM file.

        Returns:
            list[str, str, int, float]: list of tuples
            ``(gene_name, chromosome, loci, coverage)``.

        Params:
            factor (float):
                Scaling factor. Default is 2.0 (for two copies).
            regions (list[:obj:`GRange`], optional):
                List of regions to be extracted.

        Notes:
            Profiles that were used in Aldy paper:

                1. PGRNseq-v1: PGXT104 was used for all genes
                   (n.b. PGXT147 with rescale 2.52444127771 was used for CYP2B6 beta).
                2. PGRNseq-v2: NA19789.bam was used for all genes.
                3. Illumina: by definition contains all ones (uniform coverage profile).
        """
        if regions is None:
            gene_regions = sorted(
                [  # paper gene coordinates in hg19
                    ("CYP3A5", GRange("7", 99245000, 99278000)),
                    ("CYP3A4", GRange("7", 99354000, 99465000)),
                    ("CYP2C19", GRange("10", 96445000, 96615000)),
                    ("CYP2C9", GRange("10", 96691000, 96754000)),
                    ("CYP2C8", GRange("10", 96796000, 96830000)),
                    ("CYP4F2", GRange("19", 15619000, 16009500)),
                    ("CYP2A6", GRange("19", 41347500, 41400000)),
                    ("CYP2D6", GRange("22", 42518900, 42553000)),
                    ("TPMT", GRange("6", 18126541, 18157374)),
                    ("DPYD", GRange("1", 97541298, 98388615)),
                ],
                key=lambda x: x[1],
            )
        else:
            gene_regions = [(str(i), r) for i, r in enumerate(sorted(regions))]
        result: List[Tuple[str, str, int, float]] = []
        for gene, location in gene_regions:
            with pysam.AlignmentFile(sam_path) as sam:
                prefix = _chr_prefix(location.chr, sam.header["SQ"])
                region = location.samtools(pad_left=1, prefix=prefix)
                cov: dict = collections.defaultdict(
                    lambda: collections.defaultdict(int)
                )
                log.info("Generating profile for {} ({})", gene, region)
                try:
                    for read in sam.fetch(region=region):
                        start, s_start = read.reference_start, 0
                        if not read.cigartuples:
                            continue

                        for op, size in read.cigartuples:
                            if op == 2:
                                for i in range(size):
                                    cov[location.chr][start + i] += 1
                                start += size
                            elif op == 1:
                                s_start += size
                            elif op == 4:
                                s_start += size
                            elif op in [0, 7, 8]:
                                for i in range(size):
                                    cov[location.chr][start + i] += 1
                                start += size
                                s_start += size
                    result += [
                        (gene, c, p, cov[c][p] * (factor / 2.0))
                        for c in sorted(cov.keys())
                        for p in sorted(cov[c].keys())
                        if location.start - 500 <= p <= location.end + 500
                    ]
                except ValueError:
                    log.warn("Cannot fetch gene {} ({})", gene, region)
        return result


def _chr_prefix(ch: str, chrs: List[str]) -> str:  # assumes ch is not prefixed
    """
    Check if `ch` (*without any chr prefix*) should be prefixed with "chr" or not.

    Returns:
        str: Prefix to be prepended to chromosome (empty if needed).

    Params:
        ch (str): chromosome name
        chrs (list[str]): list of chromosome names in the alignment file
    """
    if ch not in chrs and "chr" + ch in chrs:
        return "chr"
    return ""


def _in_region(region: GRange, read: pysam.AlignedSegment, prefix: str) -> bool:
    """
    Returns ``True`` if a read is located within a given gene region.

    Notes:
        The region is padded with 500bp on the left side.
    """

    return (
        read.reference_id != -1
        and read.reference_name == prefix + region.chr
        and region.start - 500 <= read.reference_start <= region.end
    )



# 786
# Aldy source: coverage.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, Tuple, Callable

import collections

from .common import log
from .gene import Mutation, GeneRegion, GRange


class Coverage:
    """
    Data structure that maintains the coverage information for a given sample.
    """

    def __init__(
        self,
        coverage: Dict[int, Dict[str, int]],
        threshold: float,
        cnv_coverage: Dict[int, int],
    ) -> None:
        """
        Coverage initialization.

        Args:
            coverage (dict[int, dict[str, int]]):
                Coverage for each locus within a sample represented as a dictionary
                that maps mutation (or a reference position indicated by `_`)
                to the number of reads supporting that mutation.
                For example, ``coverage[10]['SNP.AG'] = 2`` means that there are 2
                reads that have G (instead of A) at the genomic locus 10.
            threshold (float):
                Threshold `t` used for filtering out low-quality mutations.
                Any mutation with the coverage less than `t`% is filtered out.
                Ranges from 0 to 1 (normalized percentage).
            cnv_coverage (dict[int, int]):
                Coverage of the copy-number neutral region of the sample:
                each genomic locus within that region points to the corresponsing read
                coverage.
                Used for coverage rescaling.
        """
        self._coverage = coverage
        self._threshold = threshold
        self._cnv_coverage = cnv_coverage

        self._rescaled: Dict[int, float] = {}
        self._region_coverage: Dict[Tuple[int, GeneRegion], float] = {}

    def __getitem__(self, mut: Mutation) -> float:
        """
        Returns:
            float: Coverage of the mutation ``mut``.
        """
        return self.coverage(mut)

    def coverage(self, mut: Mutation) -> float:
        """
        Returns:
            float: Coverage of the mutation ``mut``.
        """
        if mut.op in self._coverage[mut.pos]:
            return self._coverage[mut.pos][mut.op]
        else:
            return 0

    def total(self, pos: int) -> float:
        """
        Returns:
            float: Total coverage at the locus ``pos``.
        """
        if pos not in self._coverage:
            return 0
        return float(sum(v for p, v in self._coverage[pos].items() if p[:3] != "INS"))

    def percentage(self, m: Mutation) -> float:
        """
        Returns:
            float: Coverage of the mutation ``mut`` expressed as a percentage
            in the range 0-100.
        """
        total = self.total(m.pos)
        if total == 0:
            return 0
        return 100.0 * self.coverage(m) / total

    def loci_cn(self, pos: int) -> float:
        """
        Returns:
            float: Copy number of the locus ``pos``.
        """
        if self._rescaled[pos] == 0:
            return 0
        return self.total(pos) * (1 / self._rescaled[pos])

    def single_copy(self, pos: int, cn_solution) -> float:
        """
        Args:
            pos (int): genomic locus
            cn_solution (:obj:`aldy.solutions.CNSolution`): copy-number solution
        Returns:
            float: Coverage of a single gene copy at the locus ``pos``.
        """
        if cn_solution.position_cn(pos) == 0:
            return 0
        return max(1, self.total(pos)) / cn_solution.position_cn(pos)

    def region_coverage(self, gene: int, region: GeneRegion) -> float:
        """
        Returns:
            float: Average coverage of the region ``region`` in ``gene``.
        """
        return self._region_coverage[gene, region]

    def average_coverage(self) -> float:
        """
        Returns:
            float: Average coverage of the sample.
        """
        return sum(self.total(pos) for pos in self._coverage) / float(
            len(self._coverage) + 0.1
        )

    def filtered(
        self, filter_fn: Callable[[Mutation, float, float, float], bool]
    ):  # -> Coverage
        """
        Args:
            filter_fn (callable):
                Function that performs mutation filtering with the following arguments:

                    1. mut (:obj:`aldy.gene.Mutation`): mutation to be filtered
                    2. cov (float): coverage of the mutation
                    3. total (float): total coverage of the mutation locus
                    4. thres (float): filtering threshold

                ``filter_fn`` returns ``False`` if a mutation is filtered out.

        Returns:
            :obj:`Coverage`: Filtered coverage.
        """

        cov = collections.defaultdict(
            lambda: collections.defaultdict(int),
            {
                pos: collections.defaultdict(
                    int,
                    {
                        o: c
                        for o, c in pos_mut.items()
                        if filter_fn(
                            Mutation(pos, o),  # type: ignore
                            c,
                            self.total(pos),
                            self._threshold,
                        )
                    },
                )
                for pos, pos_mut in self._coverage.items()
            },
        )

        new_cov = Coverage(cov, self._threshold, self._cnv_coverage)  # type: ignore
        new_cov._rescaled = self._rescaled
        new_cov._region_coverage = self._region_coverage
        return new_cov

    def diploid_avg_coverage(self) -> float:
        """
        Returns:
            float: Average coverage of the copy-number neutral region.
        """
        return float(sum(self._cnv_coverage.values())) / abs(
            self._cn_region.end - self._cn_region.start
        )

    def _normalize_coverage(
        self,
        profile: Dict[str, Dict[int, float]],
        gene_regions: Dict[int, Dict[GeneRegion, GRange]],
        cn_region: GRange,
    ) -> None:
        """
        Normalize the sample coverage to match the profile coverage.

        Args:
            profile (dict[str, dict[int, float]]):
                Profile coverage in the form `chromosome: (position -> coverage)`.
            gene_regions
            (dict[int, dict[:obj:`aldy.common.GeneRegion`, :obj:`aldy.common.GRange`]]):
                List of genic regions for each gene.
            cn_region (:obj:`aldy.common.GRange`):
                Copy-number neutral region.
        """

        #: GRange: store the CN-neutral region
        self._cn_region: GRange = cn_region
        sam_ref = sum(
            self._cnv_coverage[i] for i in range(cn_region.start, cn_region.end)
        )
        cnv_ref = sum(
            profile[cn_region.chr][i] for i in range(cn_region.start, cn_region.end)
        )

        cn_ratio = float(cnv_ref) / sam_ref
        log.debug("CNV factor: {} ({})", cn_ratio, 1.0 / cn_ratio)

        self._rescaled = {}
        self._region_coverage = {}
        for gene, gr in gene_regions.items():
            for region, rng in gr.items():
                s = sum(self.total(i) for i in range(rng.start, rng.end))  # !IMPORTANT
                p = sum(profile[rng.chr][i] for i in range(rng.start, rng.end))
                self._rescaled.update(
                    {
                        i: profile[rng.chr][i] / cn_ratio
                        for i in range(rng.start, rng.end)
                    }
                )
                self._region_coverage[gene, region] = (
                    (cn_ratio * float(s) / p) if p != 0 else 0.0
                )

    @staticmethod
    def basic_filter(mut: Mutation, cov: float, total: float, thres: float) -> bool:
        """
        Basic filtering function.
        """
        return cov >= max(1, total * thres)

    @staticmethod
    def cn_filter(
        mut: Mutation, cov: float, total: float, thres: float, cn_solution
    ) -> bool:
        """
        Filtering function that takes into the account the copy number of the mutation.
        """
        cn = cn_solution.position_cn(mut.pos)
        total = total / cn if cn > 0 else total
        return mut.op == "_" or cov > max(1, total * thres)



# 786
# Aldy source: test_major_synthetic.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest  # noqa
import collections

from aldy.major import estimate_major
from aldy.solutions import CNSolution
from aldy.coverage import Coverage
from aldy.major import NOVEL_MUTATION_PENAL
from aldy.common import SOLUTION_PRECISION


def assert_major(gene, solver, major):
    cn_sol = CNSolution(0, list(collections.Counter(major["cn"]).elements()), gene)

    cov = collections.defaultdict(dict)
    for (pos, op), c in major["data"].items():
        cov[pos][op] = c
    cov = Coverage(cov, 0.5, {})
    sols = estimate_major(gene, cov, cn_sol, solver)

    if "score" in major:
        for s in sols:
            assert abs(major["score"] - s.score) < SOLUTION_PRECISION, "Score"
    sols_expected = [
        sorted(collections.Counter(c).elements(), key=str) for c in major["sol"]
    ]
    sols_parsed = [
        sorted(
            collections.Counter(
                {
                    tuple([k.major] + [(m[0], m[1]) for m in k.added])
                    if len(k.added) > 0
                    else k.major: v
                    for k, v in s.solution.items()
                }
            ).elements(),
            key=str,
        )
        for s in sols
    ]
    assert sorted(sols_expected) == sorted(sols_parsed)


def test_basic(toy_gene, solver):
    # Test two copies of *1
    assert_major(
        toy_gene, solver, {"cn": {"1": 2}, "data": {}, "sol": [{"1": 2}], "score": 0}
    )


def test_deletion(toy_gene, solver):
    # Test a single copy of *1C
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 1, "6": 1},
            "data": {(105, "_"): 0, (105, "SNP.TA"): 10},
            "sol": [{"1.a": 1, "6": 1}],
            "score": 0,
        },
    )


def test_two_copies(toy_gene, solver):
    # Test two copies (*1/*2)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {
                (111, "_"): 10,
                (111, "DEL.AC"): 10,
                (119, "_"): 20,
                (119, "INS.TT"): 10,
            },
            "sol": [{"1": 1, "2": 1}],
            "score": 0,
        },
    )
    # Test two copies (*2/*3)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {
                (111, "_"): 10,
                (111, "DEL.AC"): 10,
                (119, "_"): 20,
                (119, "INS.TT"): 10,
                (151, "_"): 10,
                (151, "SNP.CT"): 10,
            },
            "sol": [{"2": 1, "3": 1}],
            "score": 0,
        },
    )
    # Test slightly perturbed two copies (*2/*3)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {
                (111, "_"): 11,
                (111, "DEL.AC"): 9,
                (119, "_"): 22,
                (119, "INS.TT"): 8,
                (151, "_"): 10.5,
                (151, "SNP.CT"): 9.5,
            },
            "sol": [{"2": 1, "3": 1}],
            "score": 2 / 10 + 3 / 11 + 1 / 10,
        },
    )


def test_multiple_copies(toy_gene, solver):
    # Test four copies (*1+*1/*2+*2)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 4},
            "data": {
                (111, "_"): 20,
                (111, "DEL.AC"): 20,
                (119, "_"): 40,
                (119, "INS.TT"): 20,
            },
            "sol": [{"1": 2, "2": 2}],
            "score": 0,
        },
    )


def test_left_fusion(toy_gene, solver):
    # Test left fusion that has no SNPs (*4[*1]/*1+*2)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2, "4": 1},
            "data": {
                (111, "_"): 10,
                (111, "DEL.AC"): 10,
                (119, "_"): 20,
                (119, "INS.TT"): 10,
            },
            "sol": [{"1": 1, "2": 1, "4/1": 1}],
            "score": 0,
        },
    )
    # Test left fusion that has SNPs (*4[*3]/*1+*3)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2, "4": 1},
            "data": {(151, "_"): 10, (151, "SNP.CT"): 20},
            "sol": [{"1": 1, "3": 1, "4/3": 1}, {"3": 2, "4/1": 1}],
            "score": 0,
        },
    )
    # Test left fusion combination (*4[*3]/*1C+*3)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2, "4": 1},
            "data": {
                (105, "_"): 10,
                (105, "SNP.TA"): 10,
                (151, "_"): 10,
                (151, "SNP.CT"): 20,
            },
            "sol": [{"1.a": 1, "3": 1, "4/3": 1}],
            "score": 0,
        },
    )
    # Test left fusion combination (*4[*3]/*4[*1]+*3)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 1, "4": 2},
            "data": {
                (105, "_"): 0,
                (105, "SNP.TA"): 10,
                (151, "_"): 20,
                (151, "SNP.CT"): 10,
            },
            "sol": [{"1.a": 1, "4/3": 1, "4/1": 1}],
            "score": 0,
        },
    )


def test_right_fusion(toy_gene, solver):
    # Test right fusion (*5/*1)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 1, "5": 1},
            "data": {(111, "_"): 10, (111, "DEL.AC"): 10},
            "sol": [{"1": 1, "5": 1}],
            "score": 0,
        },
    )
    # Test right fusion (*5/*2)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 1, "5": 1},
            "data": {
                (111, "_"): 0,
                (111, "DEL.AC"): 20,
                (119, "_"): 20,
                (119, "INS.TT"): 10,
            },
            "sol": [{"2": 1, "5": 1}],
            "score": 0,
        },
    )
    # Test right fusion (*5/*3)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 1, "5": 1},
            "data": {
                (111, "_"): 10,
                (111, "DEL.AC"): 10,
                (151, "_"): 0,
                (151, "SNP.CT"): 10,
            },
            "sol": [{"3": 1, "5": 1}],
            "score": 0,
        },
    )


def test_novel_mutations(toy_gene, solver):
    # Test novel mutations within a single gene (other is deleted)
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 1, "6": 1},
            "data": {(111, "_"): 0, (111, "DEL.AC"): 20},
            "sol": [{("1", (111, "DEL.AC")): 1, "6": 1}],
            "score": NOVEL_MUTATION_PENAL,
        },
    )
    # Test novel mutations
    assert_major(
        toy_gene,
        solver,
        {
            "cn": {"1": 2},
            "data": {(111, "_"): 10, (111, "DEL.AC"): 10},
            "sol": [{"1": 1, ("1", (111, "DEL.AC")): 1}],
            "score": NOVEL_MUTATION_PENAL,
        },
    )



# 786
# Aldy source: genotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Optional, Any

import os
import sys
import textwrap

from . import sam
from . import cn
from . import major
from . import minor
from . import diplotype
from . import solutions

from .common import log, script_path, json_print, AldyException, SOLUTION_PRECISION
from .gene import Gene, GRange
from .diplotype import OUTPUT_COLS
from .lpinterface import model as lp_model


def genotype(
    gene_db: str,
    sam_path: str,
    profile: str,
    output_file: Optional[Any] = sys.stdout,
    cn_region: Optional[GRange] = None,
    cn_solution: Optional[List[str]] = None,
    threshold: float = 0.5,
    fusion_penalty: float = cn.LEFT_FUSION_PENALTY,
    solver: str = "any",
    phase: bool = False,
    reference: Optional[str] = None,
    gap: int = 0,
    max_minor_solutions: int = 1,
    debug: Optional[str] = None,
) -> List[solutions.MinorSolution]:
    """
    Genotype a sample.

    Returns:
        list[:obj:`aldy.solutions.MinorSolution`]: List of genotype solutions.

    Args:
        gene_db (str):
            Gene name (if it is located in the Aldy's gene database)
            or the location of a gene database in YML format.
        sam_path (str):
            Location of SAM/BAM/CRAM file that is to be analyzed.
        profile (str):
            Coverage profile (e.g. 'illumina').
            Can be ``None`` if ``cn_solution`` is provided.
        cn_region (:obj:`aldy.common.GRange`, optional):
            Copy-number neutral region.
            Can be ``None`` (will use the default CYP2D8 region or ``None``
            if ``cn_solution`` is provided).
        output_file (file, optional):
            Location of an output decomposition file.
            Provide ``None`` for no output.
            Default is ``sys.stdout``.
        cn_solution (list[str], optional):
            User-specified list of the copy number configurations.
            Copy-number detection solver will not run is this parameter is set.
            Default is ``None``.
        threshold (float):
            Mutation filtering threshold.
            Default is 0.5 (for 50%).
        fusion_penalty (float):
            Fusion penalty. Use higher values to avoid fusions.
            Default is 0.1.
        solver (str):
            ILP solver. Check :obj:`aldy.lpinterface` for the list of available solvers.
            Default is ``'any'``.
        phase (bool):
            Perform basic rudimentary phasing of the reads
            to aid the genotyping process.
            DEPRECATED: currently does nothing.
            Default is ``False``.
        reference (str, optional):
            A reference genome for reading CRAM files.
            Default is ``None``.
        gap (float):
            Relative optimality gap. Use non-zero values to allow non-optimal solutions.
            Default is 0 (reports only optimal solutions).
        max_minor_solutions (int):
            Maximum number of minor solutions to report for each major solution.
            Default is 1.
        debug (str, optional):
            Prefix for debug information and core dump files.
            ``None`` for no debug information.
            Default is ``None``.

    Raises:
        :obj:`aldy.common.AldyException` if the average coverage is too low (below 2).
    """

    # Test the existence of LP solver
    _ = lp_model("init", solver)

    # Load the gene specification
    db_file = script_path("aldy.resources.genes/{}.yml".format(gene_db.lower()))
    if os.path.exists(db_file):
        gene_db = db_file
    with open(gene_db):  # Check if file exists
        pass
    gene = Gene(gene_db)

    sample_name = os.path.splitext(os.path.basename(sam_path))[0]
    with open(sam_path):  # Check if file exists
        pass
    if not cn_region:
        cn_region = sam.DEFAULT_CN_NEUTRAL_REGION
    sample = sam.Sample(
        sam_path=sam_path,
        gene=gene,
        threshold=threshold,
        profile=profile,
        phase=phase,
        reference=reference,
        cn_region=cn_region,
        debug=debug,
    )

    avg_cov = sample.coverage.average_coverage()
    if avg_cov < 2:
        raise AldyException(
            f"Average coverage of {avg_cov:.2f} for gene {gene.name} is too low; "
            + f"skipping gene {gene.name}. "
            + f"Please ensure that {gene.name} is present in the input SAM/BAM."
        )
    elif avg_cov < 20:
        log.warn(
            f"Average sample coverage is {avg_cov}. "
            + "We recommend at least 20x coverage for the best results."
        )

    json_print(debug, f'"{os.path.basename(sam_path).split(".")[0]}": {{')

    # Get copy-number solutions
    json_print(debug, '  "cn": {')
    cn_sols = cn.estimate_cn(
        gene,
        sample.coverage,
        solver=solver,
        gap=gap,
        user_solution=cn_solution,
        fusion_penalty=fusion_penalty,
        debug=debug,
    )
    log.debug("*" * 80)
    json_print(debug, "  },")

    # Add SLACK to each score to avoid division by zero or other numerical issues
    # when the optimum is close to zero.
    # HACK: Value of 1 (1 gene copy) seems to be reasonable, but is not
    # theoretically guaranteed to be better than any other value.
    SLACK = 1

    # Get major solutions and pick the best one
    log.info(f"Potential {gene.name} copy number configurations for {sample_name}:")
    major_sols: list = []
    json_print(debug, '  "major": [', end="")
    cn_sols = sorted(cn_sols, key=lambda m: (int(1000 * m.score), m._solution_nice()))
    if len(cn_sols) == 0:
        raise AldyException("No solutions found!")
    min_cn_score = min(cn_sols, key=lambda m: m.score).score
    for i, cn_sol in enumerate(cn_sols):
        log.info(f"  {i + 1:2}: {cn_sol._solution_nice()}")
        conf = (min_cn_score + SLACK) / (cn_sol.score + SLACK)
        log.info(f"      Confidence: {conf:.2f} (score = {cn_sol.score:.2f})")
    log.info("")

    for i, cn_sol in enumerate(cn_sols):
        major_sols += major.estimate_major(
            gene,
            sample.coverage,
            cn_sol,
            solver=solver,
            gap=gap,
            identifier=i,
            debug=debug,
        )
    log.debug("*" * 80)
    json_print(debug, "],")
    if len(major_sols) == 0:
        raise AldyException("No major solutions found!")

    major_sols = [
        solutions.MajorSolution(
            m.score * ((m.cn_solution.score + SLACK) / (min_cn_score + SLACK)),
            m.solution,
            m.cn_solution,
        )
        for m in major_sols
    ]
    min_major_score = min(major_sols, key=lambda m: m.score).score
    major_sols = sorted(
        [m for m in major_sols if m.score - min_major_score - gap < SOLUTION_PRECISION],
        key=lambda m: (int(1000 * m.score), m._solution_nice()),
    )

    log.info(f"Potential major {gene.name} star-alleles for {sample_name}:")
    for i, major_sol in enumerate(major_sols):
        log.info(f"  {i + 1:2}: {major_sol._solution_nice()}")
        conf = (min_major_score + SLACK) / (major_sol.score + SLACK)
        log.info(f"      Confidence: {conf:.2f} (score = {major_sol.score:.2f})")
    log.info("")

    json_print(debug, '  "minor": [', end="")
    minor_sols = []
    for m in minor.estimate_minor(
        gene,
        sample.coverage,
        major_sols,
        solver,
        max_solutions=max_minor_solutions,
        debug=debug,
    ):
        n = solutions.MinorSolution(
            m.score
            * ((m.major_solution.cn_solution.score + SLACK) / (min_cn_score + SLACK)),
            m.solution,
            m.major_solution,
        )
        n.diplotype = m.diplotype
        minor_sols.append(n)

    if len(minor_sols) == 0:
        raise AldyException(
            "Aldy could not phase any major solution. "
            + "Please run with  --debug parameter and notify the authors of Aldy."
        )
    min_minor_score = min(minor_sols, key=lambda m: m.score).score
    minor_sols = sorted(
        [m for m in minor_sols if m.score - min_minor_score - gap < SOLUTION_PRECISION],
        key=lambda m: (int(1000 * m.score), m._solution_nice()),
    )
    log.debug("*" * 80)
    json_print(debug, "  ],")

    log.info(f"Best {gene.name} star-alleles for {sample_name}:")
    is_vcf = output_file and output_file.name[-4:] == ".vcf"
    if output_file and not is_vcf:
        print("#" + "\t".join(OUTPUT_COLS), file=output_file)
    for i, minor_sol in enumerate(minor_sols):
        log.info(f"  {i + 1:2}: {minor_sol.diplotype}")
        tt = textwrap.wrap(minor_sol._solution_nice(), width=60, break_long_words=False)
        t = "\n             ".join(tt)
        log.info(f"      Minor: {t}")
        conf = (min_minor_score + SLACK) / (minor_sol.score + SLACK)
        log.info(f"      Confidence: {conf:.2f} (score = {minor_sol.score:.2f})")
        if output_file and not is_vcf:
            print(f"#Solution {i + 1}: {minor_sol._solution_nice()}", file=output_file)
            diplotype.write_decomposition(
                sample_name, gene, i + 1, minor_sol, output_file
            )
    if is_vcf:
        diplotype.write_vcf(sample_name, gene, sample.coverage, minor_sols, output_file)
    json_print(debug, "},")

    return minor_sols



# 786
# Aldy source: major.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Dict, Tuple, Any, Optional

import collections
import copy

from . import lpinterface
from .common import log, AldyException, json_print, sorted_tuple, allele_sort_key
from .cn import MAX_CN
from .gene import Allele, Mutation, Gene
from .coverage import Coverage
from .solutions import CNSolution, MajorSolution, SolvedAllele


# Model parameters
NOVEL_MUTATION_PENAL = MAX_CN + 1
"""float: Penalty for each novel functional mutation (0 for no penalty).
   Should be large enough to prevent novel mutations unless really necessary."""


def estimate_major(
    gene: Gene,
    coverage: Coverage,
    cn_solution: CNSolution,
    solver: str,
    gap: float = 0,
    identifier: int = 0,
    debug: Optional[str] = None,
) -> List[MajorSolution]:
    """
    Estimate optimal major star-alleles.

    Args:
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.
        coverage (:obj:`aldy.coverage.Coverage`):
            Read coverage data.
        cn_solution (:obj:`aldy.solutions.CNSolution`):
            Copy-number solution that will be used for major star-allele calling.
        solver (str):
            ILP solver. Check :obj:`aldy.lpinterface` for the list of supported solvers.
        gap (float):
            Relative optimality gap. Use non-zero values to allow non-optimal solutions.
            Default is 0 (reports only optimal solutions).
        identifier (int):
            Unique solution identifier. Used for generating debug information.
            Default is 0.
        debug (str, optional):
            If set, create a "`debug`.major`identifier`.lp" file for debug purposes.
            Default is ``None``.

    Returns:
        list[:obj:`aldy.solutions.MajorSolution`]
    """

    log.debug("*" * 80)
    log.debug("Major solver: cn = {}", cn_solution)

    if sum(cn_solution.solution.values()) < 2:
        raise AldyException(
            "estimate_major requires at least two valid gene configurations"
        )

    alleles, coverage = _filter_alleles(gene, coverage, cn_solution)
    # Check if some CN solution has no matching allele
    if set(cn_solution.solution) - set(a.cn_config for a in alleles.values()):
        results: List[MajorSolution] = []
    else:
        results = solve_major_model(
            gene, alleles, coverage, cn_solution, solver, gap, identifier, debug
        )

    # TODO: re-implement phasing step from Aldy 1.2
    # TODO: Check for novel functional mutations and do something with them

    return results


def solve_major_model(
    gene: Gene,
    allele_dict: Dict[str, Allele],
    coverage: Coverage,
    cn_solution: CNSolution,
    solver: str,
    gap: float = 0,
    identifier: int = 0,
    debug: Optional[str] = None,
) -> List[MajorSolution]:
    """
    Solves the major star-allele detection problem via integer linear programming.

    Args:
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.
        allele_dict (dict[str, :obj:`aldy.gene.Allele`]):
            Dictionary of the candidate major star-alleles.
        coverage (:obj:`aldy.coverage.Coverage`):
            Mutation coverage data.
        cn_solution (:obj:`aldy.solutions.CNSolution`):
            Copy-number solution that will be used for calling major star-alleles
            (:obj:`aldy.solutions.CNSolution`).
        solver (str):
            ILP solver. Check :obj:`aldy.lpinterface` for the list of supported solvers.
        gap (float):
            Relative optimality gap. Use non-zero values to allow non-optimal solutions.
            Default is 0 (reports only optimal solutions).
        identifier (int):
            Unique solution identifier. Used for generating debug information.
            Default is 0.
        debug (str, optional):
            If set, create a "`debug`.major`identifier`.lp" file for debug purposes.
            Default is ``None``.

    Returns:
        list[:obj:`aldy.solutions.MajorSolution`]

    Notes:
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Major star-allele identification) for the model explanation.
    """

    model = lpinterface.model("AldyMajor", solver)

    # Get the list of _all_ functional mutations present in the sample
    # and the database (intersection)
    func_muts = {
        M for m, M in gene.mutations.items() if M.is_functional and coverage[M] > 0
    }
    _print_candidates(allele_dict, coverage, cn_solution, func_muts)

    # HACK: silence type checker
    a: Any = 0

    # Create a binary variable for every possible allele copy
    alleles = {(a, 0): allele_dict[a] for a in allele_dict}
    for (an, _), a in list(alleles.items()):
        max_cn = cn_solution.solution[a.cn_config]
        for i in range(1, max_cn):
            alleles[an, i] = alleles[an, 0]
    VA = {a: model.addVar(vtype="B", name=f"A_{a[0]}_{a[1]}") for a in alleles}

    # Make sure that VA[i+1] <= VA[i] (to avoid equivalent solutions)
    for a, ai in alleles.keys():
        if ai > 0:
            model.addConstr(VA[a, ai] <= VA[a, ai - 1], name=f"CORD_{a}_{ai}")

    # Add an error variable for every mutation
    VERR = {
        m: model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{m.pos}_{m.op}")
        for m in func_muts
    }
    constraints = {e: 0 for e in VERR}
    # Add a binary variable for each allele/novel mutation pair
    # For each binary variable, add a multiplication variable
    # MUL = VNEW[a, m] * VA[a] (binary product linearization)
    del_allele = gene.deletion_allele()
    VNEW = {
        a: {
            m: (
                model.addVar(vtype="B", name=f"N_{m.pos}_{m.op}_{a[0]}_{a[1]}"),
                model.addVar(vtype="B", name=f"MUL_N_{m.pos}_{m.op}_{a[0]}_{a[1]}"),
            )
            if a[0] != del_allele
            else (0, 0)  # deletion alleles should not be assigned any mutations
            for m in constraints
            if m[0] not in set(mm[0] for mm in alleles[a].func_muts)
            and gene.has_coverage(a[0], m.pos)
        }
        for a in alleles
    }
    # Populate the constraints
    for a in alleles:
        for m in alleles[a].func_muts:
            constraints[m] += VA[a]
    # Add novel mutation constraints
    # Also ensure that MUL = 1 <=> VA * VNEW = 1
    for a in VNEW:
        for m, (vn, vm) in VNEW[a].items():
            constraints[m] += model.prod(vm, [VA[a], vn])

    # Populate constraints of non-variations (i.e. matches with the reference genome)
    for pos in set(m.pos for m in constraints):
        ref_m = Mutation(pos, "_")  # type: ignore
        constraints[ref_m] = 0
        VERR[ref_m] = model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{pos}_REF")
        for a in alleles:
            if not gene.has_coverage(a[0], pos):
                continue
            # An insertion still contributes to the total coverage at this loci
            if any(
                ma[0] == pos and not ma[1][:3] == "INS" for ma in alleles[a].func_muts
            ):
                continue
            # Make sure that the allele has no novel mutations at this position
            constraints[ref_m] += VA[a]
            muts = [m for m in VNEW[a] if m[0] == pos and m[1][:3] != "INS"]
            for m in muts:
                constraints[ref_m] -= VNEW[a][m][1]
            # We ensure that only one additional mutation can be selected here
            model.addConstr(
                model.quicksum(VNEW[a][m][0] for m in muts) <= 1,
                name=f"CONE_{pos}_{a[0]}_{a[1]}",
            )

    # Each allele must express all of its functional mutations
    json_print(debug, "  { # {}", identifier)
    json_print(debug, f'    "cn": {str(dict(cn_solution.solution))}, ')
    json_print(debug, '    "data": {', end="")
    prev = 0
    for m, expr in sorted(constraints.items()):
        cov = coverage[m] / coverage.single_copy(m.pos, cn_solution)
        model.addConstr(expr + VERR[m] == cov, name=f"CFUNC_{m.pos}_{m.op}")
        if m.pos != prev and prev != 0:
            json_print(debug, "\n             ", end="")
        prev = m.pos
        json_print(debug, f"({m[0]}, '{m[1]}'): {cov:.4f}, ", end="")
    json_print(debug, "}, ")

    # Each CN configuration must be satisfied by corresponding alleles
    for cnf, cnt in cn_solution.solution.items():
        expr = sum(VA[a] for a in VA if alleles[a].cn_config == cnf)
        model.addConstr(expr == cnt, name=f"CSAT_{cnf}")

    # Each functional mutation must be chosen by some allele
    for m in func_muts:
        expr = model.quicksum(VA[a] for a in alleles if m in alleles[a].func_muts)
        expr += model.quicksum(
            VNEW[a][m][1]
            for a in alleles
            if m in VNEW[a] and m not in alleles[a].func_muts
        )
        model.addConstr(expr >= 1, name=f"CEXP_{m.pos}_{m.op}")

    # Objective: minimize the absolute sum of errors and the number of novel mutations
    objective = model.abssum(e for e in VERR.values())
    objective += NOVEL_MUTATION_PENAL * model.quicksum(
        VNEW[a][m][0] for a in VNEW for m in VNEW[a]
    )
    model.setObjective(objective)
    if debug:
        model.dump(f"{debug}.major{identifier}.lp")

    # Solve the model
    try:
        lookup = {
            **{model.varName(v): a for a, v in VA.items()},
            **{model.varName(v): (a, m) for a in VNEW for m, (v, _) in VNEW[a].items()},
        }
        result: Dict[Any, MajorSolution] = {}
        json_print(debug, '    "sol": [', end="")
        for status, opt, sol in model.solutions(gap):
            log.debug(f"Major solver: {status}, opt: {opt:.2f}")
            # Allele: novel mutations
            solved_alleles: Any = collections.defaultdict(lambda: [])
            for s in sol:
                if s not in lookup:
                    continue
                elif isinstance(lookup[s][0], tuple):  # Novel mutation
                    solved_alleles[lookup[s][0]].append(lookup[s][1])
                elif s not in solved_alleles:
                    solved_alleles[lookup[s]] = []
            sol_tuple = sorted_tuple(
                (a, sorted_tuple(nm)) for (a, _), nm in solved_alleles.items()
            )
            if sol_tuple not in result:
                solution = collections.Counter(
                    SolvedAllele(major=a, minor=None, added=tuple(mut), missing=tuple())
                    for (a, _), mut in solved_alleles.items()
                )
                json_print(
                    debug,
                    dict(
                        collections.Counter(
                            tuple([a] + [(m[0], m[1]) for m in mut])
                            if len(mut) > 0
                            else a
                            for (a, _), mut in solved_alleles.items()
                        )
                    ),
                    end=", ",
                )
                sol = MajorSolution(
                    score=opt, solution=solution, cn_solution=cn_solution
                )
                log.debug("Major solver: {}".format(sol))
                result[sol_tuple] = sol
        json_print(debug, "]\n  }, ", end="")
        return list(result.values())
    except lpinterface.NoSolutionsError:
        log.debug("Major solver: no solutions")
        return []


def _filter_alleles(
    gene: Gene, coverage: Coverage, cn_solution: CNSolution
) -> Tuple[Dict[str, Allele], Coverage]:
    """
    Filter out low-quality mutations and alleles that are not expressed.

    Returns:
        tuple[dict[str, :obj:`aldy.gene.Allele`], :obj:`aldy.coverage.Coverage`]:
        Tuple of allele dictionary describing the feasible alleles,
        and the coverage description of high-confidence variants.
    """

    def filter_fns(mut, cov, total, thres):
        return Coverage.basic_filter(
            mut, cov, total, thres / MAX_CN
        ) and Coverage.cn_filter(mut, cov, total, thres, cn_solution)

    cov = coverage.filtered(filter_fns)
    alleles = copy.deepcopy(gene.alleles)
    for an, a in sorted(gene.alleles.items()):
        if a.cn_config not in cn_solution.solution:
            del alleles[an]
        elif any(cov[m] <= 0 for m in a.func_muts):
            s = (
                "{} in {}".format(m, gene.region_at(m.pos))
                for m in a.func_muts
                if cov[m] <= 0
            )
            log.trace("Major solver: removing {} because of {}", an, " and ".join(s))
            del alleles[an]

    return alleles, cov


def _print_candidates(
    alleles: Dict[str, Allele],
    coverage: Coverage,
    cn_solution: CNSolution,
    func_muts: set,
) -> None:
    """
    Pretty-prints the list of allele candidates and their functional mutations.
    """
    log.debug("Major solver: candidates:")
    func_muts = func_muts.copy()
    for a in sorted(alleles, key=allele_sort_key):
        log.debug("  *{} (cn=*{})", a, alleles[a].cn_config)
        for m in sorted(alleles[a].func_muts, key=lambda m: m.pos):
            if m in func_muts:
                func_muts.remove(m)
            log.debug(
                "    {} {:4} ({:.1f} copies) {} {}",
                # coverage.region_at(m.pos),
                m,
                coverage[m],
                coverage[m] / (coverage.total(m.pos) / cn_solution.position_cn(m.pos)),
                "F",
                m.aux.get("old", ""),
            )
    if len(func_muts) > 0:
        log.debug("  Other mutations:")
        for m in sorted(func_muts, key=lambda m: m.pos):
            log.debug(
                "    {} {:4} ({:.1f} copies) {} {}",
                m,
                coverage[m],
                coverage[m] / (coverage.total(m.pos) / cn_solution.position_cn(m.pos)),
                "F",
                m.aux.get("old", ""),
            )



# 786
# Aldy source: cn.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, List, Tuple, Optional

import copy

from . import lpinterface

from .common import log, json_print, sorted_tuple, AldyException
from .gene import GeneRegion, CNConfig, Gene
from .coverage import Coverage
from .solutions import CNSolution
from .lpinterface import escape_name


PCE_REGION = GeneRegion(11, "pce")
""":obj:`aldy.common.GeneRegion`: *CYP2D7* PCE region that requires special handling
   (because *CYP2D6* does not have a matching PCE region)."""

MAX_CN = 20.0
"""float: Maximum allowed copy number."""

# Model parameters

LEFT_FUSION_PENALTY = 0.1
"""float: Extra penalty applied to left fusions to account for their rarity
   (0 for no penalty)."""

PCE_PENALTY_COEFF = 1.5
"""float: Error penalty applied to the PCE region (1 for no penalty)."""

MAX_CN_ERROR = MAX_CN
"""float: Upper bound for absolute error."""

PARSIMONY_PENALTY = 0.5
"""float: Extra penalty applied to each present gene copy to account for
   solution parsimony (0 for no penalty)."""


def estimate_cn(
    gene: Gene,
    coverage: Coverage,
    solver: str,
    gap: float = 0,
    fusion_penalty: float = LEFT_FUSION_PENALTY,
    user_solution=None,
    debug: Optional[str] = None,
) -> List[CNSolution]:
    """
    Estimate optimal copy number configurations given a gene and read data.

    Args:
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.
        coverage (:obj:`aldy.coverage.Coverage`):
            Read data coverage instance.
        solver (str):
            ILP solver. Check :obj:`aldy.lpinterface` for the list of supported solvers.
        gap (float):
            Relative optimality gap. Use non-zero values to allow non-optimal solutions.
            Default is 0 (reports only optimal solutions).
        fusion_penalty (float):
            Fusion penalty. Use higher values to avoid fusions.
            Default is 0.1.
        user_solution (list[str], optional):
            User-specified list of copy number configurations.
            ILP solver will not run if this parameter is provided.
            Default is ``None``.
        debug (str, optional):
            If set, create a "`debug`.cn.lp" file for debug purposes.
            Default is ``None``.

    Returns:
        list[:obj:`aldu.solutions.CNSolution`]: Optimal copy number configurations.
    """

    log.debug("\n" + "*" * 80)

    if user_solution is not None:
        return [_parse_user_solution(gene, user_solution)]
    else:
        # TODO: filter CN configs with non-present alleles
        max_observed_cn = 1 + max(
            int(round(coverage.region_coverage(g, r)))
            for g in gene.regions
            for r in gene.regions[g]
        )
        region_cov = _region_coverage(gene, coverage)
        total_cov = sum(r0 + r1 for r0, r1 in region_cov.values())
        min_cov = min(
            sum(sum(v.values()) for _, v in gene.cn_configs[c].cn.items())
            for c in gene.cn_configs
        )
        if total_cov < min_cov / 2.0:
            raise AldyException(
                f"Coverage for {gene.name} too low for copy number calling."
            )

        configs = _filter_configs(gene, coverage)
        log.debug("CN solver: configs = {}", ", ".join(sorted(configs)))
        log.debug("CN solver: max_cn = {}", max_observed_cn)
        _print_coverage(gene, coverage)
        sol = solve_cn_model(
            gene,
            configs,
            max_observed_cn,
            region_cov,
            solver,
            gap,
            fusion_penalty,
            debug,
        )

        return sol


def solve_cn_model(
    gene: Gene,
    cn_configs: Dict[str, CNConfig],
    max_cn: int,
    region_coverage: Dict[GeneRegion, Tuple[float, float]],
    solver: str,
    gap: float = 0,
    fusion_penalty: float = LEFT_FUSION_PENALTY,
    debug: Optional[str] = None,
) -> List[CNSolution]:
    """
    Solve the copy number estimation problem (an instance of closest vector problem).

    Args:
        cn_configs (dict[str, :obj:`aldy.gene.CNConfig`]):
            Available copy number configurations (vectors).
        max_cn (int):
            Maximum allowed copy number.
        region_coverage (dict[:obj:`aldy.common.GeneRegion`, tuple[float, float]]):
            Observed copy number of the main gene and the pseudogene
            for each genic region.
        solver (str):
            ILP solver to use.
            Check :obj:`aldy.lpinterface` for the list of supported solvers.
        gap (float):
            Relative optimality gap. Use non-zero values to allow non-optimal solutions.
            Default is 0 (reports only optimal solutions).
        fusion_penalty (float):
            Fusion penalty. Use higher values to avoid fusions.
            Default is 0.1.
        debug (str, optional):
            If set, create a "`debug`.cn.lp" file for debug purposes.
            Default is ``None``.

    Returns:
        list[:obj:`aldy.solutions.CNSolution`]: Optimal copy-number solutions.

    Notes:
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Copy number and structural variation estimation)
        for the detailed description of ILP model.
    """

    model = lpinterface.model("AldyCN", solver)

    # List of CN configurations (a.k.a. structures):
    # dict of (`name`, `number`): structure, where multiple copies of the same `name`
    # get different numbers (binary LP transformation).
    # Note that `number` = {0, -1} represents full configuration with *all* pseudogenes.
    # Thus a diploid genome must have exactly *2* full configurations.
    # Any `number` > 0 describes only the main gene configuration
    # (a.k.a. "weak" configurations), and there can be many such configurations.
    structures = {(name, 0): structure for name, structure in cn_configs.items()}
    for a, ai in list(structures.keys()):
        structures[a, -1] = copy.deepcopy(structures[a, 0])
        if cn_configs[a].kind != CNConfig.CNConfigType.DEFAULT_CN:
            continue
        for i in range(1, max_cn):
            structures[a, i] = copy.deepcopy(structures[a, 0])
            for g in structures[a, i].cn:
                # If this is a pseudogene, remove a copy
                # (i.e. create a "weak configuration")
                if g != 0:
                    structures[a, i].cn[g] = {
                        r: v - 1 for r, v in structures[a, i].cn[g].items()
                    }

    # Add a binary variable for each CN structure.
    VCN = {
        (a, ai): model.addVar(vtype="B", name=f"CN_{a}_{ai}") for a, ai in structures
    }

    # We assume diploid genome, so the number of haplotype-inducing configurations
    # must be 2.
    diplo_inducing = model.quicksum(VCN[a] for a in VCN if a[1] <= 0)
    model.addConstr(diplo_inducing == 2, name="CDIPLO")

    # Ensure that we cannot link any allele to the whole-gene deletion.
    del_allele = gene.deletion_allele()
    for (a, ai), v in VCN.items():
        if del_allele and a != del_allele:
            model.addConstr(v + VCN[del_allele, -1] <= 1, name=f"CDEL_{a}_{ai}")

    # Ensure that binary transformation is properly formed (i.e. A_i <= A_{i-1}).
    for a, ai in structures:
        # The second haplotype (-1) is only present if the first one (0) is there.
        if ai == -1:
            model.addConstr(VCN[a, ai] <= VCN[a, 0], name=f"CORD_{a}_{ai}")
        # Ignore 1, because A[1] = 1 && A[0] = 0 is valid (e.g. *13/*13+*1)
        elif ai > 1:
            model.addConstr(VCN[a, ai] <= VCN[a, ai - 1], name=f"CORD_{a}_{ai}")

    # Add error variables
    VERR = {}
    json_print(debug, '    "data": {', end="")
    for r, (exp_cov0, exp_cov1) in region_coverage.items():
        json_print(debug, f"'{str(r)[3:-1]}': ({exp_cov0}, {exp_cov1}), ", end="")
        expr = 0
        for s, structure in structures.items():
            if r in structure.cn[0]:
                expr += structure.cn[0][r] * VCN[s]
            if len(structure.cn) > 1 and r in structure.cn[1]:
                expr -= structure.cn[1][r] * VCN[s]
        VERR[r] = model.addVar(
            name="E_{}{}".format(*r), lb=-MAX_CN_ERROR, ub=MAX_CN_ERROR
        )
        model.addConstr(
            expr + VERR[r] == exp_cov0 - exp_cov1, name="CCOV_{}{}".format(*r)
        )
    json_print(debug, "},")
    # Objective: minimize the sum of absolute errors.
    # PCE_REGION (in CYP2D7) is penalized with an extra score as it is important
    # fusion marker.
    objective = model.abssum(
        VERR.values(),
        coeffs={escape_name("E_{}{}".format(*PCE_REGION)): PCE_PENALTY_COEFF},
    )
    # Objective: also minimize the total number of present alleles (maximum parsimony)
    # Also penalize left fusions as they are not likely to occur.
    objective += model.quicksum(
        # MPICL requires only one term for each variable in the objective function,
        # thus this ugly expression:
        (fusion_penalty + PARSIMONY_PENALTY) * v
        if cn_configs[s].kind == CNConfig.CNConfigType.LEFT_FUSION
        else PARSIMONY_PENALTY * v
        for (s, k), v in VCN.items()
    )
    model.setObjective(objective)
    if debug:
        model.dump(f"{debug}.cn.lp")

    # Solve the model
    try:
        lookup = {model.varName(v): a for (a, ai), v in VCN.items()}
        result: dict = {}
        for status, opt, sol in model.solutions(gap):
            log.debug(f"CN solver: {status}, opt: {opt:.2f}")
            sol_tuple = sorted_tuple(lookup[v] for v in sol)
            # Because A[1] can be 1 while A[0] is 0, we can have biologically
            # duplicate solutions
            if sol_tuple not in result:
                result[sol_tuple] = CNSolution(  # type: ignore
                    opt, solution=list(sol_tuple), gene=gene
                )
                log.debug("CN solver: {}", result[sol_tuple])
        json_print(
            debug, '    "sol": ' + str([dict(r.solution) for r in result.values()])
        )
        return list(result.values())
    except lpinterface.NoSolutionsError:
        log.debug("CN solver: no solutions")
        return []


def _filter_configs(gene: Gene, coverage: Coverage) -> Dict[str, CNConfig]:
    """
    Filter out low-quality mutations and copy number configurations
    that are not supported by the remaining mutations.

    Returns:
        dict[str, :obj:`aldy.gene.CNConfig`]
    """
    cov = coverage.filtered(
        lambda mut, cov, total, thres: Coverage.basic_filter(
            mut, cov, total, thres / MAX_CN
        )
    )
    configs = copy.deepcopy(gene.cn_configs)
    for an in sorted(gene.cn_configs):
        if an not in gene.alleles:
            continue  # This is just a CN configuration w/o any mutations
        if any(cov[m] <= 0 for m in gene.alleles[an].func_muts):
            s = (
                "{} in {}".format(m, gene.region_at(m.pos))
                for m in gene.alleles[an].func_muts
                if cov[m] <= 0
            )
            log.trace("Removing {} because of {}", an, " and ".join(s))
            del configs[an]
    return configs


def _region_coverage(
    gene: Gene, coverage: Coverage
) -> Dict[GeneRegion, Tuple[float, float]]:
    """
    Calculate the coverage  of the main gene and the pseudogene in each genic region.
    Returns dictionary where

    Returns:
        dict[:obj:`aldy.common.GeneRegion, tuple[float, float]]: Region coverage
        that links genic regions (e.g. exon 1) to the coverage of the main gene
        and the pseudogene (expressed as a tuple).
    """

    if 1 in gene.regions:  # Check if we have pseudogenes at all
        cov = {
            r: (coverage.region_coverage(0, r), coverage.region_coverage(1, r))
            for r in gene.unique_regions
            if r != PCE_REGION
        }
        # HACK: By default, CYP2D6 does not have a PCE region,
        # so we insert a dummy PCE region to prevent KeyNotFound errors
        if PCE_REGION in gene.regions[1] and PCE_REGION not in gene.regions[0]:
            cov[PCE_REGION] = (0, coverage.region_coverage(1, PCE_REGION))
    else:
        cov = {r: (coverage.region_coverage(0, r), 0) for r in gene.unique_regions}
    return cov


def _print_coverage(gene: Gene, coverage: Coverage) -> None:
    """
    Pretty-print the region coverage.
    """
    regions = set(r for g in gene.regions for r in gene.regions[g])
    log.debug("CN solver: coverage map = ")
    for r in sorted(regions):
        gc = coverage.region_coverage(0, r) if r in gene.regions[0] else 0.0
        if 1 in gene.regions:
            pc = coverage.region_coverage(1, r) if r in gene.regions[1] else 0.0
        else:
            pc = -1
        log.debug(
            "  {:5} {:2}: {:5.2f} {} {}",
            r.kind,
            r.number,
            gc,
            "" if pc == -1 else f"{pc:5.2f}",
            f"; diff = {gc - pc:5.2f}" if pc != -1 and r in gene.unique_regions else "",
        )


def _parse_user_solution(gene: Gene, sols: List[str]) -> CNSolution:
    """
    Parse user-provided copy number solutions.

    Args:
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.
        sols (list[str]):
            List of valid copy number configurations.

    Returns:
        :obj:`aldy.solutions.CNSolution`: User-provided copy number solution.

    Raises:
        :obj:`aldy.common.AldyException` if a user-provided solution does not match
        the gene database.
    """
    for sol in sols:
        if sol not in gene.cn_configs:
            raise AldyException(
                f"Given copy number solution contains unknown copy number configuration"
                + f"{sol}. Please run 'aldy show --gene {gene.name}' for the list the"
                + f"valid configurations"
            )
    s = CNSolution(0, solution=sols, gene=gene)  # type: ignore
    log.debug("CN solver: using user-provided solution: {}", s)
    return s



# 786
# Aldy source: gene.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Tuple, Dict, List, Optional

import os
import re
import enum
import yaml
import functools
import collections
import textwrap

from .common import (
    GeneRegion,
    GRange,
    AldyException,
    allele_number,
    sorted_tuple,
    seq_to_amino,
    rev_comp,
    log,
    allele_sort_key,
)


EXON = "e"
"""str: Abbreviation for exon."""

INTRON = "i"
"""str: Abbreviation for intron."""


class Allele(
    collections.namedtuple("Allele", ["name", "cn_config", "func_muts", "minors"])
):
    """
    Major allele description.
    Immutable.

    Attributes:
        name (str):
            Major allele name.
        cn_config (str):
            Copy-number configuration.
        func_muts (set[:obj:`Mutation`]):
            Functional mutations that describe the major allele.
        minors (dict[str, :obj:`Suballele`]):
            Dictionary of minor alleles that are based upon the major allele.
    """

    def minor_mutations(self, minor: str):
        """
        Sequence of mutations that describe a minor allele.
        """
        for m in self.func_muts:
            yield m
        for m in self.minors[minor].neutral_muts:
            yield m


class Suballele(
    collections.namedtuple("Suballele", ["name", "alt_names", "neutral_muts"])
):
    """
    Minor allele description.
    Immutable.

    Attributes:
        name (str):
            Minor allale name.
        alt_names (list[str]):
            List of alternative names.
        neutral_muts (set[:obj:`Mutation`]):
            Netural mutations that describe the minor allele.

    Notes:
        Has custom printer (``__repr__``).
    """

    def __str__(self):
        return "Sub({}; [{}])".format(
            "|".join([self.name] + self.alt_names),
            ", ".join(map(str, self.neutral_muts)),
        )


@functools.total_ordering
class Mutation(
    collections.namedtuple("Mutation", ["pos", "op", "is_functional", "aux"])
):
    """
    Mutation description.
    Immutable.

    Attributes:
        pos (int): Reference genome position.
        op (str): Mutation description that can be:

            - ``SNP.AB``:  a SNP from A to B (A, B are in ``[ACGT]``)
            - ``INS.xx``:  an insertion of xx (xx is ``[acgt]+``)
            - ``DEL.xx``:  a deletion of xx (xx is ``[ACGT]+``)

        is_functional (int): Functionality of the mutation:
            - 0 for non-functional (silent) mutations
            - 1 for functional (gene-disrupting) mutations
            - 2 for splice-disrupting mutations

        aux (dict[str, str]): Auxiliary information.
            Currently used for storing dbSNP rsIDs (key: ``dbsnp``)
            and old Karolinska-style mutation IDs (key: ``old``).

    Notes:
        Has custom printer (``__str__``).
        Comparable and hashable via ``(pos, op)`` tuple.
        Implements ``total_ordering``.
    """

    def __new__(self, pos: int, op: str, is_functional=0, aux=None):
        return super(Mutation, self).__new__(
            self, pos, op, is_functional, aux if aux else dict()
        )

    def __str__(self):
        return "{}:{}{}".format(self.pos, self.op, "*" if self.is_functional else "")

    def __eq__(self, other):
        return (self.pos, self.op) == (other.pos, other.op)

    def __lt__(self, other):
        return (self.pos, self.op) < (other.pos, other.op)

    def __hash__(self):
        return (self.pos, self.op).__hash__()


class CNConfig(
    collections.namedtuple("CNConfig", ["cn", "kind", "alleles", "description"])
):
    """
    Copy-number (CN) configuration description.
    Immutable.

    Attributes:
        cn (dict[int, dict[:obj:`aldy.common.GeneRegion`, int]]):
            Value of the expected region copy number in each gene.
            For example, ``cn[0][GeneRegion(0, 1, EXON)] == 1`` means that the exon 1 of
            the main gene (ID 0) has one copy (and thus should be present)
            in the configuration `cn`.
        kind (:obj:`CNConfigType`):
            Type of the copy-number configuration. See :obj:`CNConfigType` for details.
        alleles (set[str]):
            Allele IDs that have this CN configuration.
        description (str):
            Human-readable description of the configuration (e.g. "deletion").

    Notes:
        Has custom printer (``__str__``).
    """

    CNConfigType = enum.Enum(
        "CNConfigType", "DEFAULT_CN LEFT_FUSION RIGHT_FUSION DELETION"
    )
    """
    Enumeration describing the type of a copy-number configuration:

        - ``LEFT_FUSION``
        - ``RIGHT_FUSION``
        - ``DELETION``
        - ``DEFAULT_CN``
    """

    def __str__(self):
        regions = sorted(set(r for g in self.cn for r in self.cn[g]))
        return "CNConfig({}; vector={}; alleles=[{}])".format(
            str(self.kind)[13:],
            "|".join(
                "".join(
                    ("{:.0f}".format(self.cn[g][r]) if self.cn[g][r] != 0.5 else "½")
                    if r in self.cn[g]
                    else "_"
                    for r in regions
                )
                for g in sorted(self.cn)
            ),
            " ".join(sorted(self.alleles)),
        )


class Gene:
    """
    Gene (and associated pseudogenes) description.

    Attributes:
        name (str):
            Gene name (e.g. CYP2D6).
        seq (str):
            Wild-type reference sequence that describes \*1 allele.
            Should be location-compatible with hg19 (must contain no indels when
            aligned to the reference genome).
        region (:obj:`aldy.common.GRange`):
            Wild-type reference sequence coordinates within the reference genome.
        regions (dict[int, dict[:obj:`aldy.common.GeneRegion`,
                :obj:`aldy.common.GRange`]]):
            Exonic, intronic and special (e.g. UTR) regions in a gene.
            Maps a gene ID to a dictionary that maps :obj:`aldy.common.GeneRegion`
            (e.g. exon 9) to a :obj:`aldy.common.GRange` (e.g. chr1:10-20) within
            the reference genome.
            Gene 0 is the main gene.
        pseudogenes (list[str]):
            List of pseudogene names (genes whose ID is greater than 0).
        common_tandems (list[tuple[str, str]]):
            List of common allele tandems. Used in diplotype assignment heuristics.
            For example, the fact that \*13 is always followed by \*1
            (encoded as ``('13', '1')``) will be used to group \*13 and \*1 together
            within the same haplotype (e.g. \*13+\*1).
        cn_configs (dict[str, :obj:`CNConfig`]):
            Copy-number configurations associated with the gene.
            `1` (akin to \*1) is the default CN config
            (no structural variations of any kind).
        alleles (dict[str, :obj:`Allele`]):
            Major star-alleles in the gene.
            Accessible via major star-allele name (e.g. ``'1'``).
        coding_region (:obj:`CodingRegion`):
            Coding region of the gene (includes the protein sequence and
            the reference genome strand).
        mutations (dict[tuple[int, str], :obj:`Mutation`]):
            A hashtable that provides links ``(position, mutation_type)`` to
            a corresponding :obj:`Mutation`.
            Useful for querying functional mutations and auxiliary information.
    """

    CodingRegion = collections.namedtuple(
        "CodingRegion", ["rev_comp", "aminoacid", "lookup"]
    )
    """
    Immutable class that describes the coding region of a gene.
    Attributes: aminoacid sequence (``aminoacid``) and reference strand (``rev_comp``).
    """

    def __init__(
        self, path: Optional[str], name: Optional[str] = None, yml: Optional[str] = None
    ) -> None:
        """
        Initialize the Gene class with the database description
        specified in YML file ``path``.

        Args:
            path (str, optional):
                Location of YML file.
            name (str, optional):
                Gene name.

        Raises:
            :obj:`aldy.common.AldyException` if a path is not set.
        """

        if path and os.path.exists(path):
            with open(path) as file:
                yml = file.read()
            if not name:
                name = os.path.split(path)[-1].split(".")[0].upper()

        if not yml or not name:
            raise AldyException("Either a path or a name should be given")

        yml = yaml.safe_load(yml)
        self.__parse_yml(name, yml)

    def __parse_yml(self, gene_name: str, yml) -> None:
        """
        Initialize the gene structure from YML data.
        """

        self.__init_basic(gene_name, yml)
        self.__init_regions(yml)
        self.__init_alleles(yml)
        self.__init_partials()
        self.__init_structure(yml)

    def __init_basic(self, gene: str, yml) -> None:
        """
        Read basic gene properties (``name``, ``seq`` and ``region``).
        """
        self.name = gene.upper()
        self.seq = yml["seq"]
        self.region = GRange(*yml["region"])

    def __init_regions(self, yml) -> None:
        """
        Calculate the genic regions and pseudogenes
        (``regions``, ``unique_regions`` and ``pseudogenes``).
        Prepare ``region_at()`` call.
        """
        assert hasattr(self, "region")

        def calculate_regions(gene, exons, special):
            """
            Given the list of exons, calculate intronic regions and
            return a dictionary of all regions.
            """
            regions = {
                GeneRegion(e, EXON): GRange(self.region.chr, *ex)
                for e, ex in exons.items()
            }
            # Fill introns
            for i in range(1, len(regions)):
                r1, r2 = regions[GeneRegion(i + 1, EXON)], regions[GeneRegion(i, EXON)]
                if r1 > r2:
                    r1, r2 = r2, r1
                regions[GeneRegion(i, INTRON)] = GRange(self.region.chr, r1[2], r2[1])
            for n, r in special.items():
                regions[GeneRegion(n, r[0])] = GRange(self.region.chr, r[1], r[2])
            return regions

        # Gene 0 is the main gene (key is the gene ID)
        self.regions = {0: calculate_regions(0, yml["exons"], yml["special_regions"])}

        # Each pseudogene is associated with an index > 0
        self.pseudogenes: List[str] = list()
        if "pseudogenes" in yml:
            for gi, g in enumerate(sorted(yml["pseudogenes"])):
                self.pseudogenes.append(g)
                self.regions[gi + 1] = calculate_regions(
                    gi + 1,
                    yml["pseudogenes"][g]["exons"],
                    yml["pseudogenes"][g]["special_regions"],
                )

        #: dict[int, (int, `GeneRegion`]):
        #: reverse lookup (gene, region) of gene regions given a location
        #: within the reference genome.
        self._region_at = {
            i: (g, r)
            for g, d in self.regions.items()
            for r, rng in d.items()
            for i in range(rng.start, rng.end)
        }

        def parse_unique(u):
            u = list(filter(None, re.split(r"^(\d+)", u)))
            if len(u) > 1:
                return GeneRegion(int(u[0]), u[1])
            else:  # find the matching kind number (legacy support)
                for g in self.regions:
                    for r in self.regions[g]:
                        if u[0] == r.kind:
                            return r
            raise KeyError

        self.unique_regions = [parse_unique(y) for y in yml["unique_regions"]]

    def __init_alleles(self, yml) -> None:
        """
        Initialize allele (``alleles``, ``common_tandems``)
        and copy number configurations (``cn_configs``).
        Initialize old notation lookup table (``old_notation``).

        Raises:
            :obj:`aldy.common.AldyException` if allele name cannot be found.
        """

        alleles = {}
        fusions_left = {}
        fusions_right = {}

        # Human-readable CN descriptions
        descriptions = {
            "1": "Normal allelic configuration: "
            + "all regions present in both gene and pseudogene"
        }

        #: dict(:obj:`Mutation`, str):
        #: old Karolinska notation (e.g. 32C>T) for a :obj:`Mutation`.
        self._old_notation: Dict[Mutation, str] = collections.defaultdict(str)

        # allele ID of the deletion allele (i.e. whole gene is missing).
        deletion_allele = None

        for allele_name, allele in yml["alleles"].items():
            if not re.match(r"^[A-Z0-9]+\*[0-9]+[A-Z0-9]*$", allele_name):
                raise AldyException(
                    "Allele names must be in format (alphanum?)*(number)(alphanum?) "
                    + "(e.g. CYP21*2A, *3)"
                )
            allele_name = allele_name.split("*")[1]
            mutations: List[Mutation] = []
            if {"op": "deletion"} in allele["mutations"]:
                deletion_allele = allele_name
                descriptions[allele_name] = "Gene deletion"
                mutations = []
            else:
                for m in allele["mutations"]:
                    if m["pos"] == "pseudogene":  # has only one mutation indicator
                        if "-" in m["op"]:
                            fusions_left[allele_name] = (
                                int(m["op"][1:-1]),
                                (EXON if m["op"][0] == "e" else INTRON),
                            )
                            descriptions[allele_name] = (
                                "Fusion: pseudogene until "
                                + m["op"][:-1][::-1]
                                + " followed by the gene"
                            )
                        else:
                            if m["op"][-1] == "+":
                                m["op"] = m["op"][:-1]
                            fusions_right[allele_name] = (
                                int(m["op"][1:]),
                                (EXON if m["op"][0] == "e" else INTRON),
                            )
                            descriptions[allele_name] = (
                                f"Conservation: Pseudogene retention after "
                                + m["op"][::-1]
                                + " within the gene"
                            )
                    else:
                        m["aux"] = {}
                        if "dbsnp" in m:
                            m["aux"]["dbsnp"] = " or ".join(m.get("dbsnp", []))
                        if "old" in m:
                            m["aux"]["old"] = m.get("old", "")
                        mut = Mutation(m["pos"], m["op"], m["functional"], m["aux"])
                        if "old" in m:
                            self._old_notation[mut] = m["old"]
                        mutations.append(mut)
            alleles[allele_name] = Suballele(allele_name, [], mutations)

        self.common_tandems: List[tuple] = []
        if "common_tandems" in yml:
            self.common_tandems = [tuple(y) for y in yml["common_tandems"]]

        # Set copy number configurations
        # TODO: currently fusions only cover the space between the main gene and
        # the first pseudogene. Multi-pseudogene fusions are not supported.

        self.cn_configs: Dict[str, CNConfig] = dict()

        def freezekey(x):  # hashing for dictionaries
            return tuple(i[1] for i in sorted(x[0].items())) + tuple(
                i[1] for i in sorted(x[1].items())
            )

        inverse_cn: Dict[tuple, str] = dict()

        # Left fusions are PSEUDOGENE + GENE fusions
        for a, brk in fusions_left.items():
            cn = dict()
            cn[0] = {
                r: float(0 if (r.number, r.kind) < brk else 1) for r in self.regions[0]
            }
            cn[1] = {
                r: float(1 if (r.number, r.kind) < brk else 0) for r in self.regions[1]
            }

            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn, CNConfig.CNConfigType.LEFT_FUSION, {a}, descriptions[a]
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)

        # Deletion is a special kind of left fusion
        if deletion_allele is not None:
            cn = {
                0: {r: 0 for r in self.regions[0]},
                1: {r: 1 for r in self.regions[1]},
            }
            self.cn_configs[deletion_allele] = CNConfig(
                cn,
                CNConfig.CNConfigType.DELETION,
                {deletion_allele},
                descriptions[deletion_allele],
            )

        # Right fusions GENE + PSEUDOGENE + whole copy of PSEUDOGENE fusions
        for a, brk in fusions_right.items():
            cn = dict()
            cn[0] = {r: (1 if (r.number, r.kind) < brk else 0) for r in self.regions[0]}
            cn[1] = {r: (1 if (r.number, r.kind) < brk else 2) for r in self.regions[1]}
            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn, CNConfig.CNConfigType.RIGHT_FUSION, {a}, descriptions[a]
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)

        # Normal CN case
        used_alleles = {
            a for _, (_, _, alleles, _) in self.cn_configs.items() for a in alleles
        }
        has_pseudogenes = len(self.pseudogenes) > 0
        default_cn = {
            g: {r: 1 for r in self.regions[g]} for g in range(1 + has_pseudogenes)
        }

        self.cn_configs = {
            allele_number(min(v.alleles)): v for k, v in self.cn_configs.items()
        }
        self.cn_configs["1"] = CNConfig(
            default_cn,
            CNConfig.CNConfigType.DEFAULT_CN,
            set(alleles.keys()) - used_alleles,
            descriptions["1"],
        )

        # Set up major and minor allele structures
        alleles_inverse: Dict[tuple, set] = collections.defaultdict(set)
        for a in alleles:
            cn_config = next(
                cn for cn, conf in self.cn_configs.items() if a in conf.alleles
            )
            fn_muts = sorted(m for m in alleles[a].neutral_muts if m.is_functional)
            alleles_inverse[(cn_config, tuple(fn_muts))].add(a)

        self.alleles: Dict[str, Allele] = dict()
        # Karolinska DB has a lot of ambiguities, and two alleles with the same 'number'
        # can correspond to two different major alleles.
        # This step fixes this ambiguity by prepending a different letter to each unique
        # major allele that has number collision.
        used_names: Dict[str, int] = {}
        for key, minors in alleles_inverse.items():
            an = min(minors)
            a = Allele(name="", cn_config=key[0], func_muts=key[1], minors=minors)
            name = allele_number(an)
            if name in used_names:  # Append letter
                used_names[name] += 1
                name += "." + chr(used_names[name] + ord("a"))
            else:
                used_names[name] = -1
            self.alleles[name] = Allele(
                name=name,
                cn_config=a.cn_config,
                func_muts=set(a.func_muts),
                minors={
                    sa: Suballele(
                        name=sa,
                        alt_names=[],
                        neutral_muts=set(alleles[sa].neutral_muts) - set(a.func_muts),
                    )
                    for sa in a.minors
                },
            )

    def __init_partials(self) -> None:
        """
        Construct "partial" major alleles.
        If a major allele is cut in half by a fusion, we will create a "new" major
        allele that contains the functional mutations that have survived
        the fusion event.

        Note:
            This is currently supported only for left fusions.
        """

        def preserved_mutations(f, m):
            def filter_f(m):
                gene, region = self.region_at(m.pos)
                return self.cn_configs[f].cn[gene][region] > 0

            return set(filter(filter_f, m))

        for f in filter(
            lambda x: self.cn_configs[x].kind == CNConfig.CNConfigType.LEFT_FUSION,
            self.cn_configs,
        ):
            # Do not extend left fusions that are already defined
            # by some functional mutations
            # HACK: This is hacky but works well for now
            if len(self.alleles[f].func_muts) > 0:
                continue
            add: Dict[tuple, Allele] = {}
            for an, a in self.alleles.items():
                # We only make partial major alleles from non-fused major alleles
                if a.cn_config != "1":
                    continue
                new_name = "{}/{}".format(f, an)
                new_muts = preserved_mutations(f, a.func_muts)
                key = sorted_tuple(new_muts)
                if key in add:
                    add[key].minors.update(
                        {
                            san: Suballele(
                                san, [], preserved_mutations(f, sa.neutral_muts)
                            )
                            for san, sa in a.minors.items()
                        }
                    )
                else:
                    add[key] = Allele(
                        name=new_name,
                        cn_config=f,
                        func_muts=new_muts,
                        minors={
                            san: Suballele(
                                name=san,
                                alt_names=[],
                                neutral_muts=preserved_mutations(f, sa.neutral_muts),
                            )
                            for san, sa in a.minors.items()
                        },
                    )

            # Remove fusion (will be replaced at least by allele "1/{f}")
            del self.alleles[f]
            self.alleles.update({a.name: a for a in add.values()})

        for an, a in self.alleles.items():
            # Clean up minor alleles (as many might be identical after a fusion).
            # Put a reference to the cleaned-up alleles in ``alt_names`` field.
            minors: Dict[tuple, List[str]] = collections.defaultdict(list)
            for s in a.minors:
                key = sorted_tuple(a.minors[s].neutral_muts)
                minors[key].append(s)
            self.alleles[an] = Allele(
                self.alleles[an].name,
                self.alleles[an].cn_config,
                self.alleles[an].func_muts,
                {
                    min(sa): Suballele(
                        min(sa),
                        alt_names=sorted(list(set(sa) - {min(sa)})),
                        neutral_muts=set(nm),
                    )
                    for nm, sa in minors.items()
                },
            )

        # TODO: prune identical post-fusion alleles
        # (e.g. *2 after *13A and *13B is the same--- no need to have 2 items).

        for _, v in self.cn_configs.items():
            v.alleles.clear()
        for a in self.alleles.values():
            self.cn_configs[a.cn_config].alleles.add(a.name)

    def __init_structure(self, yml) -> None:
        """
        Initialize ``mutations`` lookup table
        and coding region structure ``coding_region``.
        """

        is_rev_comp = True if yml["rev_comp"] == 1 else False
        lookup = {
            i: self.seq[i - self.region.start]
            for (_, kind), (_, start, end) in self.regions[0].items()
            if kind == EXON
            for i in range(start, end)
        }
        seq = "".join(lookup[i] for i in sorted(lookup.keys()))
        aminoacid = seq_to_amino(rev_comp(seq) if is_rev_comp else seq)

        # Set up a coding region structure for aminoacid calculation
        self.coding_region = Gene.CodingRegion(is_rev_comp, aminoacid, lookup)

        self.mutations: Dict[Tuple[int, str], Mutation] = {}
        for _, a in self.alleles.items():
            self.mutations.update({(m.pos, m.op): m for m in a.func_muts})
            for _, sa in a.minors.items():
                self.mutations.update({(m.pos, m.op): m for m in sa.neutral_muts})

    # -----------------------------------------------------------------------------------

    def region_at(self, pos: int) -> Tuple[int, GeneRegion]:
        """
        Returns:
            (int, :obj:`aldy.common.GeneRegion`): Tuple consisting of a gene ID and
            a region that harbours the given position in the gene.

        Args:
            pos (int): Position in the reference genome.
        """
        return self._region_at[pos]

    def old_notation(self, mut: Mutation) -> str:
        """
        Returns:
            str: Karolinska-style notation (e.g. ``32C>T``) for a :obj:`Mutation`.

        Args:
            mut (:obj:`Mutation`)
        """
        return self._old_notation[mut]

    def sequence(self, gene: int, pad=0) -> str:
        """
        Returns:
            str: Genomic sequence of a gene.

        Args:
            id (int): Gene ID (0 for the main gene, 1+ for pseudogenes).
            pad (int): Extra padding on both sides (measured in bps). Default is 0.
        """
        regs = [(st, ed) for _, (_, st, ed) in self.regions[gene].items()]
        mi, ma = min(r[0] for r in regs), max(r[1] for r in regs)
        off = self.region.start
        return self.seq[max(0, mi - off - pad) : min(ma - off + pad, len(self.seq))]

    def is_functional(self, m: Mutation) -> bool:
        """
        Returns:
            bool: ``True`` if a mutation is functional
            (i.e. does it affect the underlying aminoacid or not).
        """
        if m.pos not in self.coding_region.lookup:
            return False
        if m.op[:3] != "SNP":
            return True
        seq = "".join(
            self.coding_region.lookup[i] if i != m.pos else m.op[5]
            for i in sorted(self.coding_region.lookup)
        )
        amino = seq_to_amino(rev_comp(seq) if self.coding_region.rev_comp else seq)
        return amino != self.coding_region.aminoacid

    def deletion_allele(self) -> Optional[str]:
        """
        Returns:
            str, optional: The deletion allele ID. Can be ``None``
            if gene has no deletion allele.
        """
        try:
            return next(
                a
                for a, cn in self.cn_configs.items()
                if cn.kind == CNConfig.CNConfigType.DELETION
            )
        except StopIteration:
            return None

    def has_coverage(self, a: str, pos: int):
        """
        Returns: ``True`` if a major allele `a` covers the mutation `m`.
        """
        m_gene, m_region = self.region_at(pos)
        return self.cn_configs[self.alleles[a].cn_config].cn[m_gene][m_region] > 0

    def print_summary(self):
        log.info(f"Gene {self.name}")
        log.info(f"  Reference genome locus: {self.region}")
        pseudo = ", ".join(f"{p} (ID {i + 1})" for i, p in enumerate(self.pseudogenes))
        log.info(f"  Pseudogenes: {pseudo}")
        amino = "\n    ".join(textwrap.wrap(self.coding_region.aminoacid))
        log.info(f"  Aminoacid:\n    {amino}")
        log.info("  Genic regions:")

        regions = set(j for _, i in self.regions.items() for j in i)
        log.info("    {:<7}  {:>25} {:>25}", "Region", "Gene", "Pseudogene")
        for r in sorted(regions):
            lg = str(self.regions[0].get(r, "-"))
            lp = str(self.regions[1].get(r, "-"))
            log.info(f"    {r.number:2} {r.kind.lower():4}: {lg:>25} {lp:>25}")

        log.info(f"  Copy number configurations:")
        a = "\n    ".join(
            textwrap.wrap(", ".join(sorted(self.cn_configs, key=allele_sort_key)))
        )
        log.info(f"    {a}")

        log.info("  Major star-alleles:")
        a = "\n    ".join(
            textwrap.wrap(
                ", ".join(f"*{a}" for a in sorted(self.alleles, key=allele_sort_key))
            )
        )
        log.info(f"    {a}")

    def print_cns(self, cn_config):
        if cn_config not in self.cn_configs:
            raise AldyException(f"Copy number configuration {cn_config} not found")
        config = self.cn_configs[cn_config]
        log.info(f"Gene {self.name}")
        log.info(f"  Copy number configuration {cn_config}:\n  {config.description}")

        alleles = sorted(config.alleles, key=allele_sort_key)
        log.info("  Major star-alleles:")
        a = "\n    ".join(textwrap.wrap(", ".join(f"*{a}" for a in alleles)))
        log.info(f"    {a}")

        log.info("  Copy number profile:")
        regions = set(j for i in config.cn for j in config.cn[i])
        log.info("    {:7}    Gene Pseudo", " ")
        for r in sorted(regions):
            lg = config.cn[0].get(r, 0.0)
            lp = config.cn[1].get(r, 0.0)
            lg = "-" if lg == 0 else f"{lg:.0f}"
            lp = "-" if lp == 0 else f"{lp:.0f}"
            log.info(f"    {r.number:2} {r.kind.lower():4}: {lg:>6} {lp:>6}")

    def print_majors(self, major):
        if major not in self.alleles:
            raise AldyException(f"Major star-allele {major} not found")
        allele = self.alleles[major]
        # Allele
        log.info(f"Gene {self.name}")
        log.info(f"Major star-allele {allele.name}:")
        log.info(f"  Copy number configuration {allele.cn_config}")

        log.info("  Functional mutations:")
        for m in sorted(allele.func_muts):
            dbsnp = m.aux.get("dbsnp", "")
            kar = m.aux.get("old", "")
            pos = f"{self.region.chr}:{m.pos}"
            log.info(f"    {pos:<11} {m.op} {dbsnp:12} {kar}")

        log.info("  Minor star-alleles:")
        a = "\n    ".join(textwrap.wrap(", ".join(f"*{a}" for a in allele.minors)))
        log.info(f"    {a}")

    def print_minors(self, minor):
        log.info(f"Gene {self.name}")
        found = False
        for major, allele in sorted(self.alleles.items()):
            if minor not in allele.minors:
                continue
            found = True
            log.info(f"\nMinor star-allele {minor}:")
            log.info(f"  Major star-allele: {allele.name}:")
            log.info(f"  Copy number configuration: {allele.cn_config}")
            if len(allele.minors[minor].alt_names) > 0:
                alt = ", ".join(sorted(allele.minors[minor].alt_names))
                log.info(f"  Alternative names: {alt}")

            log.info("  Functional mutations:")
            for m in sorted(allele.func_muts):
                dbsnp = m.aux.get("dbsnp", "")
                kar = m.aux.get("old", "")
                pos = f"{self.region.chr}:{m.pos}"
                log.info(f"    {pos:<11} {m.op} {dbsnp:12} {kar}")

            log.info("  Silent mutations:")
            for m in sorted(allele.minors[minor].neutral_muts):
                dbsnp = m.aux.get("dbsnp", "")
                kar = m.aux.get("old", "")
                pos = f"{self.region.chr}:{m.pos}"
                log.info(f"    {pos:<11} {m.op} {dbsnp:12} {kar}")
        if not found:
            raise AldyException(f"Minor star-allele {minor} not found")



# 786
# Aldy source: minor.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.

from typing import List, Set, Callable, Optional, Tuple, Dict

from . import lpinterface
from .common import log, json_print, allele_sort_key
from .gene import Mutation, Gene
from .cn import MAX_CN
from .major import NOVEL_MUTATION_PENAL
from .coverage import Coverage
from .solutions import MajorSolution, SolvedAllele, MinorSolution
from .diplotype import estimate_diplotype


# Model parameters
MISS_PENALTY_FACTOR = 1.5
"""float: Penalty for each missed minor mutation (0 for no penalty).
          Ideally larger than `ADD_PENALTY_FACTOR` as additions should be cheaper."""

ADD_PENALTY_FACTOR = 1.0
"""float: Penalty for each novel minor mutation (0 for no penalty).
          Zero penalty always prefers mutation additions over coverage errors
          if the normalized SNP slack coverage is >= 50%.
          Penalty of 1.0 prefers additions if the SNP slack coverage is >= 75%."""


def estimate_minor(
    gene: Gene,
    coverage: Coverage,
    major_sols: List[MajorSolution],
    solver: str,
    filter_fn: Optional[Callable] = None,
    max_solutions: int = 1,
    debug: Optional[str] = None,
) -> List[MinorSolution]:
    """
    Estimate the optimal minor star-allele.

    Args:
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.
        coverage (:obj:`aldy.coverage.Coverage`):
            Read coverage data.
        major_sol (:obj:`aldy.solutions.MajorSolution`):
            Major allele solution that will be used for minor star-allele calling.
        solver (str):
            ILP solver. Check :obj:`aldy.lpinterface` for the list of supported solvers.
        filter_fn (callable):
            Custom filtering function.
            Default is ``None``.
        max_solutions (int):
            Maximum number of solutions to report.
            Default is 1.
        debug (str, optional):
            If set, create a "`debug`.minor`identifier`.lp" file for debug purposes.
            Default is ``None``.

    Returns:
        list[:obj:`aldy.solutions.MinorSolution`]
    """

    # Get the list of potential alleles and mutations
    alleles: List[SolvedAllele] = list()
    # Consider all major and minor mutations
    # *from all available major solutions* together
    mutations: Set[Mutation] = set()
    for major_sol in major_sols:
        for (ma, _, added, _) in major_sol.solution:
            alleles += [
                SolvedAllele(ma, mi, added, tuple()) for mi in gene.alleles[ma].minors
            ]
            mutations |= set(gene.alleles[ma].func_muts)
            mutations |= set(added)
            for sa in gene.alleles[ma].minors.values():
                mutations |= set(sa.neutral_muts)

    # Filter out low quality mutations
    def default_filter_fn(mut, cov, total, thres):
        # TODO: is this necessary?
        if mut.op != "_" and mut not in mutations:
            return False
        return Coverage.basic_filter(
            mut, cov, total, thres / MAX_CN
        ) and Coverage.cn_filter(mut, cov, total, thres, major_sol.cn_solution)

    if filter_fn:
        cov = coverage.filtered(filter_fn)
    else:
        cov = coverage.filtered(default_filter_fn)

    minor_sols: List[MinorSolution] = []
    for i, major_sol in enumerate(
        sorted(major_sols, key=lambda s: list(s.solution.items()))
    ):
        minor_sols += solve_minor_model(
            gene, alleles, cov, major_sol, mutations, solver, i, max_solutions, debug
        )
    return minor_sols


def solve_minor_model(
    gene: Gene,
    alleles_list: List[SolvedAllele],
    coverage: Coverage,
    major_sol: MajorSolution,
    mutations: Set[Mutation],
    solver: str,
    identifier: int = 0,
    max_solutions: int = 1,
    debug: Optional[str] = None,
) -> List[MinorSolution]:
    """
    Solves the minor star-allele detection problem via integer linear programming.

    Args:
        gene (:obj:`aldy.gene.Gene`):
            Gene instance.
        alleles_list (list[:obj:`aldy.major.SolvedAllele`]):
            List of candidate minor star-alleles.
        coverage (:obj:`aldy.coverage.Coverage`):
            Sample coverage used to find out the coverage of each mutation.
        major_sol (:obj:`aldy.solutions.MajorSolution`):
            Major star-allele solution to be used for detecting minor star-alleles
            (check :obj:`aldy.solutions.MajorSolution`).
        mutations (set[:obj:`aldy.gene.Mutation`]):
            List of mutations to be considered during the solution build-up
            (all other mutations are ignored).
        solver (str):
            ILP solver to use. Check :obj:`aldy.lpinterface` for available solvers.
        identifier (int):
            Unique solution identifier. Used for generating debug information.
            Default is 0.
        max_solutions (int):
            Maximum number of solutions to report.
            Default is 1.
        debug (str, optional):
            If set, Aldy will create "<debug>.minor.lp" file for debug purposes.
            Default is ``None``.

    Returns:
        list[:obj:`aldy.solutions.MinorSolution`]

    Notes:
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Genotype refining) for the model explanation.
        Currently returns only the first optimal solution.
    """

    log.debug("*" * 80)
    log.debug("Minor solver: major = {}", major_sol)
    model = lpinterface.model("AldyMinor", solver)

    # Establish minor alleles and their mutations
    alleles: Dict[Tuple[SolvedAllele, int], Set[Mutation]] = {
        (a, 0): set(gene.alleles[a.major].func_muts)
        | set(gene.alleles[a.major].minors[a.minor].neutral_muts)
        | set(a.added)
        for a in alleles_list
    }

    _print_candidates(gene, alleles, major_sol, coverage)

    for a, _ in list(alleles):
        max_cn = major_sol.solution[SolvedAllele(a.major, None, a.added, a.missing)]
        for cnt in range(1, max_cn):
            alleles[a, cnt] = alleles[a, 0]

    VA = {
        a: model.addVar(vtype="B", name=f"A_{a[0].major}_{a[0].minor}_{a[1]}")
        for a in alleles
    }
    for a, cnt in alleles:
        if cnt > 0:
            model.addConstr(VA[a, cnt] <= VA[a, cnt - 1], name=f"CORD_{a.minor}_{cnt}")

    # Make sure that the sum of all subaleles matches the count
    # of the corresponding major alleles
    for sa, cnt in major_sol.solution.items():
        expr = model.quicksum(
            v
            for ((ma, _, ad, mi), _), v in VA.items()
            if SolvedAllele(ma, None, ad, mi) == sa
        )
        model.addConstr(expr == cnt, name=f"CCNT_{sa.major}")

    # Add a binary variable for each allele/mutation pair, where mutation belongs
    # to that allele, that will indicate whether such mutation will be kept or not.
    # Second variable stands for the corresponding VA * VKEEP binary product
    # (binary product transformation).
    VKEEP: dict = {
        a: {
            m: (
                model.addVar(
                    vtype="B", name=f"K_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}"
                ),
                model.addVar(
                    vtype="B",
                    name=f"MUL_K_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                ),
            )
            for m in alleles[a]
        }
        for a in alleles
    }
    # Add a binary variable for each allele/mutation pair, where mutation DOES NOT
    # belong to that allele,that will indicate whether such mutation will be assigned
    # to that allele or not.
    # Second variable stands for the corresponding VA * VNEW product
    # (binary product transformation).
    VNEW: dict = {
        a: {
            m: (
                model.addVar(
                    vtype="B", name=f"N_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}"
                ),
                model.addVar(
                    vtype="B",
                    name=f"MUL_N_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                ),
            )
            for m in mutations
            if gene.has_coverage(a[0].major, m.pos) and m not in alleles[a]
        }
        for a in alleles
    }
    # Add an error variable for each mutation, and populate error constraints
    VERR = {
        m: model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{m.pos}_{m.op}")
        for m in mutations
    }
    constraints = {m: 0 for m in mutations}
    for m in mutations:
        for a in alleles:
            if m in alleles[a]:
                constraints[m] += model.prod(VKEEP[a][m][1], [VA[a], VKEEP[a][m][0]])
            # Add this *only* if CN of this region in a given allele is positive
            elif gene.has_coverage(a[0].major, m.pos):
                constraints[m] += model.prod(VNEW[a][m][1], [VA[a], VNEW[a][m][0]])

    # Fill the constraints for non-variations (i.e. reference genome matches)
    for pos in set(m.pos for m in constraints):
        ref_m = Mutation(pos, "_")  # type: ignore
        VERR[ref_m] = model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{pos}_REF")
        constraints[ref_m] = 0
        for a in alleles:
            if not gene.has_coverage(a[0].major, pos):
                continue
            # Does this allele contain any mutation at the position `pos`?
            # Insertions are not counted as they always contribute to `_`.
            present_muts = [m for m in alleles[a] if m.pos == pos and m[1][:3] != "INS"]
            assert len(present_muts) < 2
            if len(present_muts) == 1:
                constraints[ref_m] += VA[a] - VKEEP[a][present_muts[0]][1]
            else:
                constraints[ref_m] += VA[a]
                muts = [m for m in VNEW[a] if m.pos == pos and m[1][:3] != "INS"]
                for m in muts:
                    constraints[ref_m] -= VNEW[a][m][1]
                # We ensure that only one additional mutation can be selected here
                model.addConstr(
                    model.quicksum(VNEW[a][m][0] for m in muts) <= 1,
                    name=f"CONE_{pos}_{a[0].major}_{a[0].minor}_{a[1]}",
                )

    # Ensure that each constraint matches the observed coverage
    json_print(debug, "  {")
    json_print(debug, f'    "cn": {str(dict(major_sol.cn_solution.solution))}, ')
    json_print(
        debug,
        '    "major": {}'.format(
            {
                tuple([s.major] + [(m[0], m[1]) for m in s.added])
                if len(s.added) > 0
                else s.major: v
                for s, v in major_sol.solution.items()
            }
        ),
        end=", ",
    )
    json_print(debug, '    "data": {', end="")
    prev = 0
    for m, expr in sorted(constraints.items()):
        scov = coverage.single_copy(m.pos, major_sol.cn_solution)
        # If scov = 0, no mutations should be selected at that locus
        # (enforced by other constraints)
        cov = coverage[m] / scov if scov > 0 else 0
        model.addConstr(expr + VERR[m] == cov, name=f"CCOV_{m.pos}_{m.op}")
        if m.pos != prev and prev != 0:
            json_print(debug, "\n             ", end="")
        prev = m.pos
        json_print(debug, f"({m.pos}, '{m.op}'): {coverage[m]:4}, ", end="")
    json_print(debug, "}, ")

    # Enforce the following rules:
    # 1) Each mutation is assigned only to alleles that are present in the solution
    for a, mv in VKEEP.items():
        for m, v in mv.items():
            model.addConstr(
                v[0] <= VA[a],
                name=f"CVK_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
            )
    for a, mv in VNEW.items():
        for m, v in mv.items():
            model.addConstr(
                v[0] <= VA[a],
                name=f"CVN_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
            )
    # 2) Each allele must express ALL of its functional mutations
    for a in alleles:
        for m in alleles[a]:
            if m.is_functional:
                assert m in VKEEP[a]
                model.addConstr(
                    VKEEP[a][m][0] >= VA[a],
                    name=f"CFUNC_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 3) No allele can include a mutation with coverage 0
    for a in VKEEP:
        for m, v in VKEEP[a].items():
            if not gene.has_coverage(a[0].major, m.pos):
                model.addConstr(
                    v[0] <= 0,
                    name=f"CZERO_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 4) No allele can include an extra functional mutation
    #    (that should be done in the major model)
    for a in VNEW:
        for m, v in VNEW[a].items():
            if m.is_functional:
                model.addConstr(
                    v[0] <= 0,
                    name=f"CNEWFUNC_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 5) Avoid extra mutations if there is an existing mutation at the corresponding
    #    locus (either from the definition or added via VNEW)
    for pos in set(m.pos for m in constraints):
        for a in alleles:
            mp = [VKEEP[a][m][1] for m in VKEEP[a] if m.pos == pos]
            ma = [VNEW[a][m][1] for m in VNEW[a] if m.pos == pos]
            # TODO: add support for extra insertions!
            if len(ma) > 1:  # useless if == 1 since variables are binary
                model.addConstr(
                    model.quicksum(ma) <= 1,
                    name=f"CSINGLE_{pos}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
            if len(ma) + len(mp) > 1:
                model.addConstr(
                    model.quicksum(mp + ma) <= 1,
                    name=f"CSINGLEFULL_{pos}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 5) Make sure that each mutation copy has at least one read supporting it
    for m in mutations:
        expr = model.quicksum(VKEEP[a][m][1] for a in alleles if m in VKEEP[a])
        expr += model.quicksum(VNEW[a][m][1] for a in alleles if m in VNEW[a])
        if major_sol.cn_solution.position_cn(m.pos) == 0 or coverage[m] == 0:
            model.addConstr(expr <= 0, name=f"CNOCOV_{m.pos}_{m.op}")
        else:
            model.addConstr(expr <= coverage[m], name=f"CMAXCOV_{m.pos}_{m.op}")
            # Ensure that at least one allele picks an existing non-filtered mutation
            model.addConstr(expr >= 1, name=f"CMINONE_{m.pos}_{m.op}")
    # 6) Do the same for non-mutations
    for pos in set(m.pos for m in constraints):
        expr = 0
        for a in alleles:
            e = [VKEEP[a][m][1] for m in VKEEP[a] if m.pos == pos]
            e += [VNEW[a][m][1] for m in VNEW[a] if m.pos == pos]
            expr += len(e) * VA[a] - model.quicksum(e)
        if isinstance(expr, int) and expr == 0:
            continue
        m = Mutation(pos, "_")
        if major_sol.cn_solution.position_cn(m.pos) == 0:
            model.addConstr(expr <= 0, name=f"CNOCOV_{m.pos}_{m.op}")
        else:
            # If there is no coverage at `pos` at all despite CN being > 0 (or if there
            # is coverage only on a functional mutation that cannot be selected),
            # allow the allele to select a non-existent non-mutation to prevent
            # infeasibility. Should not happen with "sane" datasets...
            model.addConstr(
                expr <= max(major_sol.cn_solution.position_cn(m.pos), coverage[m]),
                name=f"CMAXCOV_{m.pos}_{m.op}",
            )

    # Objective: minimize the absolute sum of errors ...
    objective = model.abssum(v for _, v in VERR.items())
    # ... and penalize the misses ...
    objective += MISS_PENALTY_FACTOR * model.quicksum(
        len(VKEEP[a]) * VA[a] for a in VKEEP
    )
    objective -= MISS_PENALTY_FACTOR * model.quicksum(
        v[1] for a in VKEEP for _, v in VKEEP[a].items()
    )
    # ... and additions ...
    objective += ADD_PENALTY_FACTOR * model.quicksum(
        v[0] for a in VNEW for _, v in VNEW[a].items()
    )
    # ... and novel functional mutations from the major model!
    objective += NOVEL_MUTATION_PENAL * model.quicksum(
        v[0]
        for a in VKEEP
        for m, v in VKEEP[a].items()
        if m.is_functional and m not in gene.alleles[a[0].major].func_muts
    )

    model.setObjective(objective)
    if debug:
        model.dump(f"{debug}.minor{identifier}.lp")

    # Solve the model
    try:
        results = {}
        for status, opt, sol in model.solutions():
            log.debug(f"Minor solver: {status}, opt: {opt:.2f}")

            solution = []
            for allele, value in VA.items():
                if model.getValue(value) <= 0:
                    continue
                added: List[Mutation] = []
                missing: List[Mutation] = []
                for m, mv in VKEEP[allele].items():
                    if not model.getValue(mv[0]):
                        missing.append(m)
                for m, mv in VNEW[allele].items():
                    if model.getValue(mv[0]):
                        added.append(m)
                solution.append(
                    SolvedAllele(
                        allele[0].major,
                        allele[0].minor,
                        allele[0].added + tuple(added),
                        tuple(missing),
                    )
                )
            json_print(
                debug,
                '    "sol": {}, '.format(
                    [
                        (
                            s.minor,
                            [(m[0], m[1]) for m in s.added],
                            [(m[0], m[1]) for m in s.missing],
                        )
                        for s in solution
                    ]
                ),
            )
            sol = MinorSolution(score=opt, solution=solution, major_solution=major_sol)
            _ = estimate_diplotype(gene, sol)
            json_print(debug, f'    "diplotype": "{sol.diplotype}"')
            json_print(debug, "  },")
            log.debug(f"Minor solver: {sol}")
            if str(sol) not in results:
                results[str(sol)] = sol
            if len(results) >= max_solutions:
                break
        return results.values()
    except lpinterface.NoSolutionsError:
        log.debug("Minor solver: no solutions")
        json_print(debug, '    "sol": []')
        json_print(debug, "  },")
        if debug and False:  # Enable to debug infeasible models
            model.model.computeIIS()
            model.dump(f"{debug}.iis.ilp")
        return []


def _print_candidates(gene, alleles, major_sol, coverage):
    """
    Pretty-prints the list of allele candidates and their mutations.
    """

    log.debug("Minor solver: candidates:")
    for a in sorted(alleles, key=lambda x: allele_sort_key(x[0].minor)):
        (ma, mi, _, _), _ = a
        log.debug("  *{} (cn=*{})", mi, gene.alleles[ma].cn_config)
        for m in sorted(alleles[a], key=lambda m: m.pos):
            m_gene, m_region = gene.region_at(m.pos)
            scopy = coverage.single_copy(m.pos, major_sol.cn_solution)
            log.debug(
                "    {:26}  {:.2f} ({:4} / {} * {:4.0f}) {}:{:10} {:>20}",
                str(m),
                coverage[m] / scopy if scopy > 0 else 0,
                coverage[m],
                major_sol.cn_solution.position_cn(m.pos),
                coverage.single_copy(m.pos, major_sol.cn_solution),
                m_gene,
                str(m_region),
                m.aux.get("old", ""),
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

exec(open("aldy/version.py").read())

setup(
    name="aldy",
    version=__version__,
    description="A tool for allelic decomposition (haplotype reconstruction) "
    + "and exact genotyping of highly polymorphic and structurally variant genes",
    url="http://aldy.csail.mit.edu/",
    author="Ibrahim Numanagić",
    author_email="inumanag@mit.edu",
    download_url="https://github.com/inumanag/aldy/tarball/master",
    license="Aldy/IURTC License. Free for academic/non-commercial use.",
    keywords=["cyp2d6", "adme", "genotyping", "illumina", "pgrnseq", "getrm", "allele"],
    install_requires=["pyyaml", "logbook", "pysam", "pytest", "ortools"],
    entry_points={"console_scripts": ["aldy = aldy.__main__:main"]},
    packages=find_packages(),
    package_data={
        "aldy.resources": ["*.rst", "aldy/resources/*.rst"],
        "aldy.resources.genes": ["*.yml", "aldy/resources/genes/*.yml"],
        "aldy.resources.profiles": ["*.profile", "aldy/resources/profiles/*.profile"],
        "aldy.tests": ["*.yml", "aldy/tests/resources/*.yml"],
        "aldy.tests.resources": [
            "*.json",
            "aldy/tests/resources/*.json",
            "*.bai",
            "aldy/tests/resources/*.bai",
            "*.bam",
            "aldy/tests/resources/*.bam",
            "*.dump",
            "aldy/tests/resources/*.dump",
            "*.expected",
            "aldy/tests/resources/*.expected",
        ],
    },
    test_suite="pytest-runner",
    tests_require=["pytest"],
)
